import glob
import os
import shutil
import warnings
from pathlib import Path  # start moving from os & glob to pathlib

from loguru import logger as log
from stubber.utils.repos import match_lib_with_mpy

from .. import utils

# Classes and functions from makemanifest to ensure that the manifest.py files can be processed
from .common import get_freeze_path, get_portboard

# globals
FAMILY = "micropython"


def freeze_folders(stub_folder: str, mpy_folder: str, lib_folder: str, version: str):
    """
    get and parse the to-be-frozen .py modules for micropython to extract the static type information
    locates the to-be-frozen files in modules folders
    - 'ports/<port>/modules/*.py'
    - 'ports/<port>/boards/<board>/modules/*.py'
    """
    match_lib_with_mpy(version_tag=version, lib_path=Path(lib_folder))

    targets = []
    scripts = glob.glob(mpy_folder + "/ports/**/modules/*.py", recursive=True)
    if len(scripts) > 0:
        # clean target folder
        shutil.rmtree(stub_folder, ignore_errors=True)
    for script in scripts:
        port, board = get_portboard(Path(script))

        freeze_path, board = get_freeze_path(Path(stub_folder), port, board)
        dest_path = freeze_path.as_posix()
        # if board == "":
        #     board = "GENERIC"

        # dest_path = os.path.join(stub_folder, port, board)

        log.debug("freeze_internal : {:<30} to {}".format(script, dest_path))
        # ensure folder, including possible path prefix for script
        os.makedirs(dest_path, exist_ok=True)
        # copy file
        try:
            shutil.copy2(script, dest_path)
            if not dest_path in targets:
                targets.append(dest_path)
        except OSError as e:
            ## Ignore errors that are caused by reorganisation of Micropython-lib
            # log.exception(e)
            warnings.warn("unable to freeze {} due to error {}".format(e.filename, str(e)))

    for dest_path in targets:
        # make a module manifest
        port = dest_path.split("/")[-2]
        # todo: add board / variant into manifest files ?
        utils.make_manifest(
            Path(dest_path),
            family=FAMILY,
            port=port,
            version=version,
            stubtype="frozen",
        )
    return targets



"""
typed-config-toml

Extend typed-config to read configuration from .toml files

"""
# TODO : extend support for . notation in section names

from typing import Dict, Optional
from pathlib import Path

try:
    import tomllib  # type: ignore
except ModuleNotFoundError:
    import tomli as tomllib

from typedconfig.config import Config, key, section# type: ignore
from typedconfig.source import ConfigSource, EnvironmentConfigSource# type: ignore

from loguru import logger as log


class TomlConfigSource(ConfigSource):
    """Read configuration from a .toml file

    prefix is used to allow for toml nested configuration
    a common prefix = "tool."

    ```
    #pyproject.toml
    [tool.deadparrot]
    species = "Norwegian Blue"
    state = "resting"
    details = ["pinging","Lovely plumage","3"]
    ```
    The use the below code to retrieve:
    ```
    # TODO
    ```
    """

    def __init__(self, filename: str, prefix: Optional[str] = None, must_exist: bool = True):
        self.filename = filename
        toml_dict = {}
        if Path(self.filename).exists():
            # Read data - will raise an exception if problem with file
            try:
                with open(filename, "rb") as f:
                    toml_dict = tomllib.load(f)
            except tomllib.TOMLDecodeError:  # pragma: no cover
                log.warning("unable to read ")
        elif must_exist:
            raise FileNotFoundError(f"Could not find config file {self.filename}")
        if not prefix or len(prefix) == 0:
            self.data = toml_dict
        else:
            # common prefix = "tool."
            try:
                self.data = toml_dict[prefix.rstrip(".")]
            except KeyError:  # pragma: no cover
                self.data = {}

        # Quick checks on data format

        assert isinstance(self.data, Dict)
        for k, v in self.data.items():
            assert isinstance(k, str)
            assert isinstance(v, Dict)
            for v_k, v_v in v.items(): # type: ignore
                assert isinstance(v_k, str)
                # do not assume/require that all values are strings
                # assert isinstance(v_v, str)
        # Convert all keys to lowercase
        self.data = {k.lower(): {v_k.lower(): v_v for v_k, v_v in v.items()} for k, v in self.data.items()}

    def get_config_value(self, section_name: str, key_name: str) -> Optional[str]:
        # Extract info from data which we read in during __init__
        section = self.data.get(section_name.lower(), None)
        if section is None:
            return None
        return section.get(key_name.lower(), None)



# type: ignore
from .config import readconfig
from .manifest import make_manifest, manifest
from .post import do_post_processing
from .stubmaker import generate_pyi_files, generate_pyi_from_file
from .versions import clean_version



"""
Download or update the micropyton compatibility modules from pycopy and stores them in the all_stubs folder
The all_stubs folder should be mapped/symlinked to the micropython_stubs/stubs repo/folder
"""
# pragma: no cover
import json
import os
import pkgutil
import shutil
import subprocess
import tempfile
from pathlib import Path

from loguru import logger as log

from . import __version__, utils
from .utils.config import CONFIG

# # log = logging.getLogger(__name__)


def get_core(requirements, stub_path=None, family: str = "core"):
    "Download MicroPython compatibility modules"
    if not stub_path:
        stub_path = CONFIG.stub_path / "cpython-core"  # pragma: no cover
    stub_path = Path(stub_path)

    # use pip to dowload requirements file to build folder in one go
    #   pip install --no-compile --no-cache-dir --target ./scratch/test --upgrade -r ./src/micropython.txt

    os.makedirs(stub_path, exist_ok=True)
    # get a temp dir and clean up afterwards
    with tempfile.TemporaryDirectory() as tmpdir:
        build_path = Path(tmpdir).absolute()
        # os.makedirs(build_path, exist_ok=True)
        mod_manifest = None
        modlist = []

        # within package/mymodule1.py, for example

        data = pkgutil.get_data(__name__, "data/" + requirements)
        if not data:  # pragma: no cover
            raise Exception("Resource Not found")
        temp_file = tempfile.NamedTemporaryFile(prefix="requirements", suffix=".txt", delete=False)
        with temp_file.file as fp:
            fp.write(data)
            fp.close
        #
        req_filename = temp_file.name

        try:
            subprocess.run(
                [
                    "pip",
                    "install",
                    "--target",
                    build_path.as_posix(),
                    "-r",
                    req_filename,
                    "--no-cache-dir",
                    "--no-compile",
                    "--upgrade",
                    "--no-binary=:all:",
                ],
                capture_output=False,
                check=True,
            )

        except OSError as err:  # pragma: no cover
            log.error(
                "An error occurred while trying to run pip to download the MicroPython compatibility modules from PyPi: {}".format(err)
            )

        # copy *.py files in build folder to stub_path
        # sort by filename to reduce churn in the repo
        for filename in sorted(build_path.rglob("*.py")):
            log.debug("pipped : {}".format(filename.name))
            # f_name, f_ext = os.path.splitext(os.path.basename(filename))  # pylint: disable=unused-variable
            modlist.append({"file": filename.name, "module": filename.stem})
            try:
                shutil.copy2(filename, stub_path)
            except OSError as err:  # pragma: no cover
                log.exception(err)

    # build modules.json
    mod_manifest = utils.manifest(family="cpython-core", port=family, version=__version__, stubtype="core", platform="cpython")
    mod_manifest["modules"] += modlist

    if mod_manifest and len(modlist):
        # write the the module manifest for the cpython core modules
        with open(stub_path / "modules.json", "w") as outfile:
            json.dump(mod_manifest, outfile, indent=4, sort_keys=True)


if __name__ == "__main__":
    # just run a quick test
    # logging.basicConfig(format="%(levelname)-8s:%(message)s", level=logging.INFO)
    get_core(requirements="requirements-core-pycopy.txt", stub_path="./scratch/cpython_common")



"""
prepare a set of stub files for publishing to PyPi

required folder structure:
NOTE: stubs and publish paths can be located in different locations and repositories

+--stubs                                                               - [config.stubs_path]
|  +--<any stub folders in repo>
|  +--micropython-v1_18-esp32


+--publish                                                             - [config.publish_path]
|  +--package_data.jsondb
|  +--package_data_test.jsondb
|  +--template                                                         - [config.template_path]
|     +--pyproject.toml
|     +--README.md
|     +--LICENSE.md
|  +--<folder for each package>
|     +--<package name> double nested to match the folder structure
|  +--<family>-<port>-<board>-<type>-stubs-<version>
|  +--micropython-esp32-stubs-v1_18
|  +--micropython-stm32-stubs-v1_18
|  +--micropython-stm32-stubs-v1_19_1
|  +-- ...
|


!!Note: anything excluded in .gitignore is not packaged by poetry
"""

from itertools import chain
from typing import Any, Dict, List, Union

from loguru import logger as log
from pysondb import PysonDB
from stubber.publish.candidates import frozen_candidates
from stubber.publish.database import get_database
from stubber.publish.package import (StubSource, create_package,
                                     get_package_info, package_name)
from stubber.publish.pypi import Version, get_pypy_versions
from stubber.publish.stubpacker import StubPackage
from stubber.utils.config import CONFIG
from stubber.utils.versions import clean_version


# ######################################
# micropython-doc-stubs
# ######################################
# todo : Publish: Integrate doc stubs in publishing loop
def publish(
    *,
    db: PysonDB,
    pkg_type,
    version: str,
    family="micropython",
    production,  # PyPI or Test-PyPi
    dryrun=False,  # don't publish , dont save to the database
    force=False,  # publish even if no changes
    clean: bool = False,  # clean up afterards
    port: str = "",
    board: str = "",
) -> Dict[str, Any]:
    """
    Publish a package to PyPi
    look up the previous package version in the dabase, and only publish if there are changes to the package
    - change determied by hash across all files

    """
    # semver, no prefix
    version = clean_version(version, drop_v=True, flat=False)
    # package name for firmware package
    pkg_name = package_name(pkg_type=pkg_type, port=port, board=board, family=family)
    status: Dict[str, Any] = {"result": "-", "name": pkg_name, "version": version, "error": None}
    log.debug("=" * 40)

    package_info = get_package_info(
        db,
        CONFIG.publish_path,
        pkg_name=pkg_name,
        mpy_version=version,
    )
    if package_info:
        # create package from the information retrieved from the database
        package = StubPackage(pkg_name, version=version, json_data=package_info)

    else:
        log.debug(f"No package found for {pkg_name} in database, creating new package")
        package = create_package(
            pkg_name,
            mpy_version=version,
            port=port,
            board=board,
            family=family,
            pkg_type=pkg_type,
        )
    log.info(f"Processing {package.package_path.name}")
    log.trace(f"{package.package_path.as_posix()}")
    status["version"] = package.pkg_version
    # check if the sources exist
    ok = True
    for (name, path) in package.stub_sources:
        if not (CONFIG.stub_path / path).exists():
            msg = f"{pkg_name}: source '{name}' not found: {CONFIG.stub_path / path}"
            if not name == StubSource.FROZEN:
                log.warning(msg)
                log.warning(msg)
                status["error"] = msg
                ok = False
            else:
                # not a blocking issue if there are no frozen stubs, perhaps this port/board does not have any
                log.warning(msg)
    if not ok:
        log.warning(f"{pkg_name}: skipping as one or more source stub folders are missing")
        package._publish = False
        # TODO Save ?
        return status
    try:
        package.update_package_files()
        package.update_included_stubs()
        package.check()
    except Exception as e:  # pragma: no cover
        log.error(f"{pkg_name}: {e}")
        status["error"] = str(e)
        return status

    # If there are changes to the package, then publish it
    if not (package.is_changed() or force):
        log.debug(f"No changes to package : {package.package_name} {package.pkg_version}")
    else:
        if not force:  # pragma: no cover
            log.info(f"Found changes to package : {package.package_name} {package.pkg_version}")
            log.debug(f"Old hash {package.hash} != New hash {package.create_hash()}")
        ## TODO: get last published version.postXXX from PyPI and update version if different
        if not dryrun:
            # only bump version if we are going to publish
            # try to get version from PyPi and increase past that
            old_ver = package.pkg_version
            pypi_versions = get_pypy_versions(package.package_name, production=production, base=Version(package.pkg_version))
            if pypi_versions:
                package.pkg_version = str(pypi_versions[-1])
            # to get the next version
            new_ver = package.bump()

            log.debug(f"{pkg_name}: bump version for {old_ver} to {new_ver} {production}")
        # Update hashes
        package.update_hashes()
        package.write_package_json()

        status["version"] = package.pkg_version

        log.trace(f"New hash: {package.package_name} {package.pkg_version} {package.hash}")
        result = package.build()
        if not result:
            log.warning(f"{pkg_name}: skipping as build failed")
            status["error"] = "Build failed"
            return status

        if dryrun:  # pragma: no cover
            log.warning("Dryrun: Updated package is NOT published.")
            status["result"] = "DryRun successful"
        else:
            result = package.publish(production=production)
            if not result:
                log.warning(f"{pkg_name}: Publish failed for {package.pkg_version}")
                status["error"] = "Publish failed"
                return status
            status["result"] = "Published"
            db.add(package.to_dict())
            db.commit()
            # TODO: push to github
            # git add <Publish folder>
            # git commit -m "Publish micropython-esp32-stubs (1.18.post24)"
            # git push
            # add tag ?

    if clean:
        package.clean()
    return status


def publish_one(
    family="micropython",
    versions: Union[str, List[str]] = "v1.18",
    ports: Union[str, List[str]] = "auto",
    boards: Union[str, List[str]] = "GENERIC",
    frozen: bool = False,
    production=False,
    dryrun: bool = False,
    clean: bool = False,
    force: bool = False,
):
    "Publish a bunch of stub packages"
    db = get_database(CONFIG.publish_path, production=production)
    l = list(frozen_candidates(family=family, versions=versions, ports=ports, boards=boards))
    result = None
    if len(l) > 0:
        todo = l[0]
        result = publish(
            db=db,
            dryrun=dryrun,
            clean=clean,
            force=force,
            production=production,
            **todo,
        )
    return result


def publish_multiple(
    family="micropython",
    versions: Union[str, List[str]] = ["v1.18", "v1.19"],
    ports: Union[str, List[str]] = "auto",
    boards: Union[str, List[str]] = "GENERIC",
    frozen: bool = False,
    production=False,
    dryrun: bool = False,
    clean: bool = False,
    force: bool = False,
):
    "Publish a bunch of stub packages"
    db = get_database(CONFIG.publish_path, production=production)

    worklist = []
    results = []
    if frozen:
        worklist += list(
            chain(
                frozen_candidates(family=family, versions=versions, ports=ports, boards=boards),
                # frozen_candidates(family="micropython", versions="v1.19.1", ports="auto", boards="auto"),
            )
        )
    for todo in worklist:

        result = publish(
            db=db,
            dryrun=dryrun,
            clean=clean,
            force=force,
            production=production,
            **todo,
        )
        results.append(result)

    return results



import json
import os
from pathlib import Path
from typing import Optional

from .my_version import __version__
from .versions import clean_version

# # log = logging.getLogger(__name__)
# # logging.basicConfig(level=logging.INFO)


def manifest(
    family: str = "micropython",
    stubtype: str = "frozen",
    machine: Optional[str] = None,  # also frozen.variant
    port: Optional[str] = None,
    platform: Optional[str] = None,
    sysname: Optional[str] = None,
    nodename: Optional[str] = None,
    version: Optional[str] = None,
    release: Optional[str] = None,
    firmware: Optional[str] = None,
) -> dict:
    "create a new empty manifest dict"

    machine = machine or family  # family
    port = port or "common"  # family
    platform = platform or port  # family
    version = version or "0.0.0"
    sysname = sysname or ""
    nodename = nodename or sysname or ""
    release = release or version or ""
    if firmware is None:
        firmware = "{}-{}-{}".format(family, port, clean_version(version, flat=True))
        # remove double dashes x2
        firmware = firmware.replace("--", "-")
        firmware = firmware.replace("--", "-")

    mod_manifest = {
        "$schema": "https://raw.githubusercontent.com/Josverl/micropython-stubber/main/data/schema/stubber-v1_4_0.json",
        "firmware": {
            "family": family,
            "port": port,
            "platform": platform,
            "machine": machine,
            "firmware": firmware,
            "nodename": nodename,
            "version": version,
            "release": release,
            "sysname": sysname,
        },
        "stubber": {
            "version": __version__,
            "stubtype": stubtype,
        },
        "modules": [],
    }
    return mod_manifest


def make_manifest(folder: Path, family: str, port: str, version: str, release: str = "", stubtype: str = "", board: str = "") -> bool:
    """Create a `module.json` manifest listing all files/stubs in this folder and subfolders."""
    mod_manifest = manifest(family=family, port=port, machine=board, sysname=family, version=version, release=release, stubtype=stubtype)
    try:
        # list all *.py files, not strictly modules but decent enough for documentation
        files = list(folder.glob("**/*.py"))
        if len(files) == 0:
            files = list(folder.glob("**/*.pyi"))

        # sort the list
        for file in sorted(files):
            # if file is in folder, then use relative path only
            # use old style relative path determination to support # python 3.8
            file = Path(os.path.relpath(file, start=folder))
            # if file.is_relative_to(folder):
            #     file = file.relative_to(folder)

            mod_manifest["modules"].append(
                {
                    "file": file.as_posix(),
                    "module": file.stem,
                }
            )

        # write the the module manifest
        with open(folder / "modules.json", "w") as outfile:
            json.dump(mod_manifest, outfile, indent=4)
        return True
    except OSError:
        return False



#!/usr/bin/env python3
"""
Collect modules and python stubs from the Loboris MicroPython source project and stores them in the all_stubs folder
The all_stubs folder should be mapped/symlinked to the micropython_stubs/stubs repo/folder
"""

# pylint: disable= line-too-long
# Copyright (c) 2020 Jos Verlinde
# MIT license
from pathlib import Path
from typing import Optional

from . import downloader, utils
from .utils.config import CONFIG

FAMILY = "loboris"
PORT = "esp32_lobo"


def get_frozen(stub_path: Optional[Path] = None, *, repo: Optional[str] = None, version="3.2.24"):
    "Download Loboris frozen modules direct from github repo"
    if stub_path is None:
        stub_path = CONFIG.stub_path / "{}-{}-frozen".format(FAMILY, utils.clean_version(version, flat=True))
    else:
        stub_path = Path(stub_path)

    if not repo:
        repo = "https://raw.githubusercontent.com/loboris/MicroPython_ESP32_psRAM_LoBo/master/MicroPython_BUILD/components/micropython/esp32/modules/{}"

    frozen_modules = [
        "README.md",
        "ak8963.py",
        "freesans20.py",
        "functools.py",
        "logging.py",
        "microWebSocket.py",
        "microWebSrv.py",
        "microWebTemplate.py",
        "mpu6500.py",
        "mpu9250.py",
        "pye.py",
        "ssd1306.py",
        "tpcalib.py",
        #        "upip.py",
        #        "upip_utarfile.py",
        #        "upysh.py",
        "urequests.py",
        "writer.py",
    ]
    # download
    downloader.download_files(repo, frozen_modules, stub_path)
    # make a manifest
    utils.make_manifest(
        stub_path,
        FAMILY,
        port="esp32",
        version=version,
        stubtype="frozen",
    )



from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

import libcst as cst


@dataclass
class TypeInfo:
    "contains the  functiondefs and classdefs info read from the stubs source"
    name: str
    decorators: Sequence[cst.Decorator]
    params: Optional[cst.Parameters] = None
    returns: Optional[cst.Annotation] = None
    docstr_node: Optional[cst.SimpleStatementLine] = None
    def_node: Optional[Union[cst.FunctionDef, cst.ClassDef]] = None
    def_type: str = "?"  # funcdef or classdef or module


class TransformError(Exception):
    """
    Error raised upon encountering a known error while attempting to transform
    the tree.
    """


MODULE_KEY = tuple(["__module"])

# debug helper
_m = cst.parse_module("")


class StubTypingCollector(cst.CSTVisitor):
    def __init__(self):
        # stack for storing the canonical name of the current function
        self.stack: List[str] = []
        # store the annotations
        self.annotations: Dict[
            Tuple[str, ...],  # key: tuple of canonical class/function name
            TypeInfo,  # value: TypeInfo
        ] = {}

    # ------------------------------------------------------------
    def visit_Module(self, node: cst.Module) -> bool:
        "Store the module docstring"
        if node.get_docstring():
            ## TODO: try / catch
            assert isinstance(node.body[0], cst.SimpleStatementLine)
            ti = TypeInfo(
                name="module",
                params=None,
                returns=None,
                docstr_node=node.body[0],
                decorators=(),
                def_type="module",
            )
            self.annotations[MODULE_KEY] = ti
        return True

    # ------------------------------------------------------------
    #  keep track of the the (class, method) names to the stack
    def visit_ClassDef(self, node: cst.ClassDef) -> Optional[bool]:
        "Store the class docstring"
        self.stack.append(node.name.value)
        if node.get_docstring():
            ## TODO: try / catch
            assert isinstance(node.body.body[0], cst.SimpleStatementLine)
            docstr_node = node.body.body[0]
        else:
            docstr_node = None

        ti = TypeInfo(
            name=node.name.value,
            params=None,
            returns=None,
            docstr_node=docstr_node,
            decorators=node.decorators,
            def_type="classdef",
            def_node=node,
        )
        self.annotations[tuple(self.stack)] = ti

    def leave_ClassDef(self, original_node: cst.ClassDef) -> None:
        self.stack.pop()

    # ------------------------------------------------------------
    def visit_FunctionDef(self, node: cst.FunctionDef) -> Optional[bool]:
        "store each function/method signature"
        self.stack.append(node.name.value)
        if node.get_docstring():
            ## TODO: try / catch
            assert isinstance(node.body.body[0], cst.SimpleStatementLine)
            docstr_node = node.body.body[0]
        else:
            docstr_node = None

        ti = TypeInfo(
            name=node.name.value,
            params=node.params,
            returns=node.returns,
            docstr_node=docstr_node,
            decorators=node.decorators,
            def_type="funcdef",
            def_node=node,
        )
        self.annotations[tuple(self.stack)] = ti

    def leave_FunctionDef(self, original_node: cst.FunctionDef) -> None:
        self.stack.pop()


def update_def_docstr(
    dest_node: Union[cst.FunctionDef, cst.ClassDef],
    src_comment: Optional[cst.SimpleStatementLine],
    src_node=None,
) -> Any:
    """
    Update the docstring of a function/method or class

    for functiondefs ending in an ellipsis, the entire body needs to be replaced.
    in this case the src_body is mandatory.
    """
    if not src_comment:
        return dest_node

    # function def on a single line ending with an ellipsis (...)
    if isinstance(dest_node.body, cst.SimpleStatementSuite):
        # in order to add a boy the simple hack is to copy the src_node.body
        if src_node:
            return dest_node.with_changes(body=src_node.body)
        else:
            return dest_node

    # just checking
    if not isinstance(dest_node.body, cst.IndentedBlock):
        raise TransformError("Expected Def with Indented body")
    # if not isinstance(dest_node.body.body, Sequence):
    #     # this is likely a .pyi file or a type declaration with a trailing ...
    #     # no changes
    #     return dest_node

    # classdef of functiondef with an indented body
    # need some funcky casting to avoid issues with changing the body
    # note : indented body is nested : body.body
    if dest_node.get_docstring() != None:
        body = tuple([src_comment] + list(dest_node.body.body[1:]))
    else:
        # append the new docstring and append the function body
        body = tuple([src_comment] + list(dest_node.body.body))

    body_2 = dest_node.body.with_changes(body=body)

    return dest_node.with_changes(body=body_2)


def update_module_docstr(node: cst.Module, doc_tree: Optional[cst.SimpleStatementLine]) -> Any:
    "Update the docstring of a module"
    if not doc_tree:
        return node
    # # just checking
    # if not (isinstance(node.body, Sequence)):
    #     raise TransformError("Expected module with body")

    # need some funcky casting to avoid issues with changing the body
    if node.get_docstring() != None:
        body = tuple([doc_tree] + list(node.body[1:]))
    else:
        # append the new docstring and append the function body
        body = tuple([doc_tree] + list(node.body))

    return node.with_changes(body=body)



"""
enrich machinestubs with docstubs
"""
from pathlib import Path
from typing import Union

import click
from loguru import logger as log
from stubber.codemod.enrich import enrich_folder
from stubber.utils.config import CONFIG

from .cli import stubber_cli


@stubber_cli.command(name="enrich")
@click.option(
    "--stubs",
    "-s",
    "stubs_folder",
    default=CONFIG.stub_path.as_posix(),
    type=click.Path(exists=True, file_okay=False, dir_okay=True),
    help="folder containing the firmware stubs to be updated",
    show_default=True,
)
@click.option(
    "--docstubs",
    "-ds",
    "docstubs_folder",
    default=CONFIG.stub_path.as_posix(),
    type=click.Path(exists=True, file_okay=False, dir_okay=True),
    help="folder containing the docstubs to be applied",
    show_default=True,
)
@click.option("--diff", default=False, help="Show diff", show_default=True, is_flag=True)
@click.option("--dry-run", default=False, help="Dry run does not write the files back", show_default=True, is_flag=True)
def cli_enrich_folder(
    stubs_folder: Union[str, Path],
    docstubs_folder: Union[str, Path],
    diff=False,
    dry_run=False,
):
    """
    Enrich the stubs in stub_folder with the docstubs in docstubs_folder.
    """
    write_back = not dry_run
    log.info(f"Enriching {stubs_folder} with {docstubs_folder}")
    _ = enrich_folder(Path(stubs_folder), Path(docstubs_folder), show_diff=diff, write_back=write_back, require_docstub=False)



"""
prepare a set of stub files for publishing to PyPi

"""

import sys
from pathlib import Path
from typing import Dict, List, Tuple, Union

from loguru import logger as log
from packaging.version import parse
from pysondb import PysonDB
from stubber.publish.enums import COMBO_STUBS, CORE_STUBS, DOC_STUBS, StubSource
from stubber.publish.stubpacker import StubPackage
from stubber.utils.versions import clean_version

# replace std log handler with a custom one capped on INFO level
log.remove()
log.add(sys.stderr, level="INFO", backtrace=True, diagnose=True)


def package_name(pkg_type, port: str = "", board: str = "", family="micropython", **kwargs) -> str:
    "generate a package name for the given package type"
    if pkg_type == COMBO_STUBS:
        # # {family}-{port}-{board}-stubs
        return f"{family}-{port}-{board}-stubs".lower().replace("-generic-stubs", "-stubs")
    elif pkg_type == DOC_STUBS:
        return f"{family}-doc-stubs".lower()
    elif pkg_type == CORE_STUBS:
        return f"{family}-core-stubs".lower()

    raise NotImplementedError(port, board, pkg_type)


# def package_path(port, board, mpy_version, pub_path: Path, pkg=COMBINED, family="micropython") -> Path:
#     "generate a package name"
#     return pub_path / package_name( port, board, pkg, family)


def get_package_info(db: PysonDB, pub_path: Path, *, pkg_name: str, mpy_version: str) -> Union[Dict, None]:
    """
    get a package's record from the json db if it can be found
    matches om the package name and version
        pkg_name: package name (micropython-esp32-stubs)
        mpy_version: micropython/firmware version (1.18)
    """
    # find in the database
    recs = db.get_by_query(query=lambda x: x["mpy_version"] == mpy_version and x["name"] == pkg_name)
    # dict to list
    recs = [{"id": key, "data": recs[key]} for key in recs]
    # sort
    packages = sorted(recs, key=lambda x: parse(x["data"]["pkg_version"]))

    # [
    #     log.trace(
    #         f"{x['data']['name']} - {x['data']['mpy_version']} - {x['data']['pkg_version']}"
    #     )
    #     for x in packages
    # ]
    if len(packages) > 0:
        pkg_from_db = packages[-1]["data"]
        log.debug(f"Found latest {pkg_name} == {pkg_from_db['pkg_version']}")
        return pkg_from_db
    else:
        return None


def create_package(
    pkg_name: str,
    mpy_version: str,
    *,
    port: str = "",
    board: str = "",
    family: str = "micropython",
    pkg_type=COMBO_STUBS,
) -> StubPackage:
    """
    create and initialize a package with the correct sources

    """
    package = None
    if pkg_type == COMBO_STUBS:
        assert port != ""
        assert board != ""
        ver_flat = clean_version(mpy_version, flat=True)
        stubs: List[Tuple[str, Path]] = [
            (
                StubSource.FIRMWARE,
                # TODO: look for the most specific firmware stub folder that is available ?
                # is it possible to prefer micropython-nrf-microbit-stubs over micropython-nrf-stubs
                # that would also require the port - board - variant to be discoverable runtime
                Path(f"{family}-{ver_flat}-{port}"),
            ),
            (
                StubSource.FROZEN,
                Path(f"{family}-{ver_flat}-frozen") / port / board,
            ),
            (
                StubSource.CORE,
                Path("cpython_core-pycopy"),
            ),
        ]
        package = StubPackage(pkg_name, version=mpy_version, stubs=stubs)
    elif pkg_type == DOC_STUBS:
        # TODO add doc stubs
        ver_flat = clean_version(mpy_version, flat=True)

        stubs: List[Tuple[str, Path]] = [
            (
                "Doc stubs",
                Path(f"{family}-{ver_flat}-docstubs"),
            ),
        ]
        package = StubPackage(pkg_name, version=mpy_version, stubs=stubs)

    elif pkg_type == CORE_STUBS:
        # TODO add core stubs
        raise NotImplementedError(type)
    else:
        raise NotImplementedError(type)

    return package



from pathlib import Path

from .typed_config_toml import (Config, EnvironmentConfigSource,
                                TomlConfigSource, key, section)


@section("micropython-stubber")
class StubberConfig(Config):
    "stubber confguration class"
    stub_path = key(key_name="stub-path", cast=Path, required=False, default=Path("./stubs"))
    "a Path to the stubs directory"
    # relative to stubs folder
    fallback_path = key(key_name="fallback-path", cast=Path, required=False, default=Path("typings/fallback"))
    "a Path to the fallback stubs directory"

    # ------------------------------------------------------------------------------------------
    # micropython and micropython-lib are relative to ./repo folder
    repo_path = key(key_name="repo-path", cast=Path, required=False, default=Path("./repos"))
    "a Path to the repo directory"

    mpy_path = key(key_name="mpy-path", cast=Path, required=False, default=Path("micropython"))
    "a Path to the micropython folder in the repos directory"

    mpy_lib_path = key(key_name="mpy-lib-path", cast=Path, required=False, default=Path("micropython-lib"))
    "a Path to the micropython-lib folder in the repos directory"

    # mpy_stubs_repo_path = key(key_name="mpy-stubs-repo-path", cast=Path, required=False, default=Path("./micropython-stubs"))
    # "a Path to the micropython-stubs folder in the repos directory"

    publish_path = key(key_name="publish-path", cast=Path, required=False, default=Path("./repos/micropython-stubs/publish"))
    "a Path to the folder where all stub publication artefacts are stored"

    template_path = key(key_name="template-path", cast=Path, required=False, default=Path("./repos/micropython-stubs/publish/template"))
    "a Path to the publication folder that has the template files"

    def post_read_hook(self) -> dict:
        config_updates = dict()
        # relative to stubs
        config_updates.update(fallback_path=self.stub_path / self.fallback_path)

        # relative to repo path
        config_updates.update(mpy_path=self.repo_path / self.mpy_path)
        config_updates.update(mpy_lib_path=self.repo_path / self.mpy_lib_path)
        # config_updates.update(mpy_stubs_repo_path=self.repo_path / self.mpy_stubs_repo_path)
        return config_updates


def readconfig(filename: str = "pyproject.toml", prefix: str = "tool.", must_exist: bool = True):
    "read the configuration from the pyproject.toml file"
    # locate the pyproject.toml file
    path = Path.cwd()
    while not (path / filename).exists():
        path = path.parent
        if path == Path("/"):
            raise FileNotFoundError(f"Could not find config file {filename}")
    filename = str(path / filename)

    config = StubberConfig()
    # add provider sources to the config
    config.add_source(EnvironmentConfigSource())
    config.add_source(TomlConfigSource(filename, prefix=prefix, must_exist=must_exist))  # ,"tools.micropython-stubber"))
    config.read()
    return config


#######################
# config singleton
CONFIG = readconfig()
"stubber configuration singleton"
#######################



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
import argparse
from typing import List
from libcst import Module, Comment
from libcst.codemod import CodemodContext, VisitorBasedCodemodCommand
from libcst.helpers.module import insert_header_comments

import re
from typing import Pattern


class AddComment(VisitorBasedCodemodCommand):
    DESCRIPTION: str = "Add comment(s) to each file"

    @staticmethod
    def add_args(arg_parser: argparse.ArgumentParser) -> None:
        # Add command-line args that a user can specify for running this
        # codemod.
        arg_parser.add_argument(
            "-c",
            "--comment",
            dest="comments",
            action="extend",
            nargs="+",
            metavar="COMMENT",
            help="Comment statement(s) to add to each file",
            type=str,
            required=True,
        )

    def __init__(self, context: CodemodContext, comments: List[str]) -> None:
        # Initialize the base class with context, and save our args. Remember, the
        # "dest" for each argument we added above must match a parameter name in
        # this init.
        super().__init__(context)
        self.comments = [c if c[0] == "#" else f"# {c}" for c in comments]
        # regex only checks for the first line of the comment
        self._regex_pattern: Pattern[str] = re.compile(rf"^{self.comments[0]}\s*$")
        self.needs_add = True

    def visit_Comment(self, node: Comment) -> None:
        if self._regex_pattern.search(node.value):
            self.needs_add = False

    def leave_Module(self, original_node: Module, updated_node: Module) -> Module:
        # If the tag already exists, don't modify the file.
        if not self.needs_add:
            return updated_node
        return insert_header_comments(updated_node, self.comments)



#!/usr/bin/env python3
"""
Collect modules and python stubs from MicroPython source projects (v1.12 +) and stores them in the all_stubs folder
The all_stubs folder should be mapped/symlinked to the micropython_stubs/stubs repo/folder

"""

# Copyright (c) 2020 Jos Verlinde
# MIT license
# some functions used from micropython/micropython/tools/makemanifest.py,
#   part of the MicroPython project, http://micropython.org/
#   Copyright (c) 2019 Damien P. George

# locating frozen modules :
# tested on MicroPython v1.12 - v1.13
# - 1.16 - using manifests.py, include can specify kwargs
# - 1.13 - using manifests.py, and support for variant
# - 1.12 - using manifests.py, possible also include content of /port/modules folder ?
# - 1.11 and older - include content of /port/modules folder if it exists
import os
import shutil  # start moving from os & glob to pathlib
from pathlib import Path
from typing import List, Optional

from loguru import logger as log
from packaging.version import Version
from stubber import utils
from stubber.freeze.freeze_folder import freeze_folders  # Micropython < v1.12
from stubber.freeze.freeze_manifest_1 import \
    freeze_one_manifest_1  # Micropython v1.12 - 1.19.1
from stubber.freeze.freeze_manifest_2 import freeze_one_manifest_2
from stubber.utils.config import CONFIG

# globals
FAMILY = "micropython"


def get_manifests(mpy_path: Path) -> List[Path]:
    """return a list of all manifests.py files found in the ports folder"""
    log.info(f"looking for manifests in  {mpy_path}")
    all_manifests = [
        m.absolute()
        for m in (mpy_path / "ports").rglob("manifest.py")
        if Path(m).parent.name != "coverage" and not "venv" in m.parts and not ".venv" in m.parts
    ]
    log.info(f"manifests found: {len(all_manifests)}")
    return all_manifests


def freeze_any(stub_folder: Path, version: str, mpy_path: Optional[Path] = None, mpy_lib_path: Optional[Path] = None):  #
    """
    get and parse the to-be-frozen .py modules for micropython to extract the static type information
     - requires that the MicroPython and Micropython-lib repos are checked out and available on a local path
     - repos should be cloned side-by-side as some of the manifests refer to micropython-lib scripts using a relative path

    the repos should be checked out to the version

    """
    count = 0
    current_dir = os.getcwd()
    if not mpy_path:
        mpy_path = CONFIG.mpy_path.absolute()
    else:
        mpy_path = Path(mpy_path).absolute()
    if not mpy_lib_path:
        mpy_lib_path = CONFIG.mpy_path.absolute()
    else:
        mpy_lib_path = Path(mpy_lib_path).absolute()

    if not stub_folder:
        frozen_stub_path = Path("{}/{}_{}_frozen".format(CONFIG.stub_path, FAMILY, utils.clean_version(version, flat=True))).absolute()
    else:
        frozen_stub_path: Path = Path(stub_folder).absolute()

    # if old version of micropython, use the old freeze method
    if version not in ["master", "latest"] and Version(version) <= Version("1.11"):
        log.debug("MicroPython v1.11, older or other")
        # others
        modules = freeze_folders(frozen_stub_path.as_posix(), mpy_path.as_posix(), mpy_lib_path.as_posix(), version)
        count = len(modules)
    else:
        # get the manifests of the different ports and boards
        all_manifests = get_manifests(mpy_path)

        # process all_manifests under the ports folder and update the frozen files in the stubs folder
        # we are goning to jump around, avoid relative paths
        frozen_stub_path = frozen_stub_path.absolute()
        mpy_path = mpy_path.absolute()
        mpy_lib_path = mpy_lib_path.absolute()

        if len(all_manifests) > 0:
            log.info(f"manifests: {len(all_manifests)}")
            shutil.rmtree(frozen_stub_path, ignore_errors=True)
        else:
            log.warning("no manifests found")
        for manifest in all_manifests:
            # TODO: try processing older version with new algoritm (1.12 - 1.19.1)
            if version in ["latest", "master"] or Version(version) >= Version("1.12"):
                try:
                    freeze_one_manifest_2(manifest, frozen_stub_path, mpy_path, mpy_lib_path, version)
                    count += 1
                except Exception as e:
                    log.warning(f"Failed to freeze manifest {manifest} with manifest_2, trying manifest_1")
                    try:
                        freeze_one_manifest_1(manifest, frozen_stub_path, mpy_path, mpy_lib_path, version)
                        count += 1
                    except Exception as e:
                        log.error(f"Error processing manifest {manifest} : {e}")

            elif Version(version) >= Version("1.12"):
                try:
                    freeze_one_manifest_1(manifest, frozen_stub_path, mpy_path, mpy_lib_path, version)
                    count += 1
                except Exception as e:
                    log.error(f"Error processing manifest {manifest} : {e}")
    # restore cwd
    os.chdir(current_dir)
    return count



"""
find the version of this package
"""

try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:  # pragma: no cover
    import importlib_metadata  # type: ignore

__version__ = "0.0.0"
try:
    # todo: use __name__
    __version__ = importlib_metadata.version("micropython-stubber")
except Exception:  # pragma: no cover
    __version__ = "0.0.0"



##########################################################################################
# get-frozen
##########################################################################################
from pathlib import Path
from typing import List

import click
import stubber.basicgit as git
from stubber.freeze.get_frozen import freeze_any
import stubber.utils as utils
from loguru import logger as log
from stubber.utils.config import CONFIG

from .cli import stubber_cli

##########################################################################################


@stubber_cli.command(name="get-frozen")
@click.option(
    "--stub-folder",
    "-stubs",
    default=CONFIG.stub_path.as_posix(),
    type=click.Path(exists=True, file_okay=False, dir_okay=True),
    show_default=True,
)
# @click.option("--path", "-p", default=config.repo_path.as_posix(), type=click.Path(file_okay=False, dir_okay=True), show_default=True)
@click.option("--version", "--tag", default="", type=str, help="Version number to use. [default: Git tag]")
@click.option("--pyi/--no-pyi", default=True, help="Create .pyi files for the (new) frozen modules", show_default=True)
@click.option("--black/--no-black", default=True, help="Run black on the (new) frozen modules", show_default=True)
def cli_get_frozen(
    stub_folder: str = CONFIG.stub_path.as_posix(),
    # path: str = config.repo_path.as_posix(),
    version: str = "",
    pyi: bool = True,
    black: bool = True,
):
    """
    Get the frozen stubs for MicroPython.

    Get the frozen modules for the checked out version of MicroPython
    """

    stub_paths: List[Path] = []

    if len(version) == 0:
        version = utils.clean_version(git.get_tag(CONFIG.mpy_path.as_posix()) or "0.0")
    if version:
        log.info("MicroPython version : {}".format(version))
        # folder/{family}-{version}-frozen
        family = "micropython"
        stub_path = Path(stub_folder) / f"{family}-{utils.clean_version(version, flat=True)}-frozen"
        stub_paths.append(stub_path)
        freeze_any(stub_path, version=version, mpy_path=CONFIG.mpy_path, mpy_lib_path=CONFIG.mpy_lib_path)
    else:
        log.warning("Unable to find the micropython repo in folder : {}".format(CONFIG.mpy_path.as_posix()))
    log.info(f"::group:: start post processing of retrieved stubs")
    utils.do_post_processing(stub_paths, pyi, black)
    log.info(f"::group:: Done")



"""
 Processing for createstubs.py
 minimizes and cross-compiles a micropyton file.
"""
import itertools
import subprocess
from pathlib import Path
from typing import Tuple, Union, List


from loguru import logger as log

try:
    import python_minifier
except ImportError:  # pragma: no cover
    python_minifier = None


def edit_lines(content, edits, diff=False):
    """Edit string by list of edits

    Args:
        content (str): content to edit
        edits ([(str, str)]): List of edits to make.
            The first string in the tuple representsthe type of edit to make, can be either:
            - comment - comment text out (removed on minify)
            - rprint - replace text with print
            - rpass - replace text with pass
            The second string is the matching text to replace
        diff (bool, optional): Prints diff of each edit.
            Defaults to False.

    Returns:
        str: edited string
    """

    def comment(l, x):
        return l.replace(x, f"# {x}")

    def rprint(l, x): # type: ignore # lgtm [py/unused-local-variable] pylint: disable= unused-variable 
        split = l.split("(")
        if len(split) > 1:
            return l.replace(split[0].strip(), "print")
        return l.replace(x, f"print")

    def rpass(l, x):  # type: ignore # lgtm [py/unused-local-variable] pylint: disable= unused-variable 
        return l.replace(x, f"pass")

    def get_whitespace_context(content, index):
        """Get whitespace count of lines surrounding index"""

        def count_ws(line):
            return sum(1 for _ in itertools.takewhile(str.isspace, line))

        lines = content[index - 1 : index + 2]
        context = (count_ws(l) for l in lines)
        return context

    def handle_multiline(content, index):
        """Handles edits that require multiline comments

        Example:
            self._log.debug("info: {} {}".format(
                1,
                2
            ))
        Here, only commenting out the first self._log line will raise
        an error. So this function returns all lines that need to
        be commented out instead.

        It also checks for situations such as this:
            if condition:
                self._log.debug('message')

        Here, since the only functionality of the conditional is the call log,
        both lines would be returned to comment out.

        """
        line = content[index]
        open_cnt = line.count("(")
        close_cnt = line.count(")")
        ahead_index = 1
        look_ahead = 0
        while not open_cnt == close_cnt:  # pragma: no cover
            look_ahead = l_index + ahead_index
            ahead_index += 1
            next_l = content[look_ahead]
            open_cnt += next_l.count("(")
            close_cnt += next_l.count(")")
        if ahead_index > 1:  # pragma: no cover
            return range(index, look_ahead + 1)
        prev = content[index - 1]
        _, line_ws, post_ws = get_whitespace_context(content, index)
        prev_words = prev.strip().strip(":").split()
        check = any(
            t
            in (
                "if",
                "else",
            )
            for t in prev_words
        )
        if check and line_ws != post_ws:
            return range(index - 1, index + 1)

    def handle_try_except(content, index):
        """Checks if line at index is in try/except block

        Handles situations like this:
            try:
                something()
            except:
                self._log.debug('some message')

        Simply removing the self._log call would create a syntax error,
        which is what this function checks for.

        """
        prev = content[index - 1]
        _, line_ws, post_ws = get_whitespace_context(content, index)
        if "except" in prev and line_ws != post_ws:
            return True

    lines = []
    multilines = set()
    content = content.splitlines(keepends=True)
    for line in content:
        _line = line
        for edit, text in edits:
            if text in line:
                if edit == "comment":
                    l_index = content.index(line)
                    # Check if edit spans multiple lines
                    mline = handle_multiline(content, l_index)
                    if mline:
                        multilines.update(mline)
                        break
                    # Check if line is only statement in try/except
                    if handle_try_except(content, l_index):
                        edit = "rpass"
                        text = line.strip()
                func = eval(edit)  # pylint: disable= eval-used
                line = func(line, text)
                if line != _line:
                    if diff:
                        print(f"\n- {_line.strip()}")
                        print(f"+ {line.strip()}")
                    break
        lines.append(line)
    for line_num in multilines:
        # Go back and comment out multilines
        line = lines[line_num]
        lines[line_num] = comment(line, line.strip())
    stripped = "".join(lines)
    return stripped


def minify_script(source_script: Path, keep_report=True, diff=False) -> str:
    """minifies createstubs.py

    Args:
        keep_report (bool, optional): Keeps single report line in createstubs
            Defaults to True.
        diff (bool, optional): Print diff from edits. Defaults to False.

    Returns:
        str: minified source text
    """

    edits:List[Tuple[str,str]] = [
        ("comment", "print"),
        ("comment", "import logging"),
        # first full
        ("comment", "self._log ="),
        ("comment", "self._log("),
        ("comment", "self._log.debug"),
        ("comment", "self._log.warning"),
        ("comment", "self._log.info"),
        ("comment", "self._log.error"),
        # then short versions
        ("comment", "_log ="),
        ("comment", "_log.debug"),
        ("comment", "_log.warning"),
        ("comment", "_log.info"),
        ("comment", "_log.error"),
    ]
    if keep_report:
        report = (
            "rprint",
            ('self._log.info("Stub module: {:<20} to file:' ' {:<55} mem:{:>5}".' "format(module_name, file_name, m1))"),
        )
        clean = (
            "rprint",
            'self._log.info("Clean/remove files in folder: {}".format(path))',
        )
        edits.insert(0, report)
        edits.insert(1, clean)

    with source_script.open("r") as f:
        if not python_minifier:  # pragma: no cover
            raise Exception("python_minifier not available")

        content = f.read()

        content = edit_lines(content, edits, diff=diff)

        source = python_minifier.minify(
            content,
            filename=source_script.name,
            combine_imports=True,
            remove_literal_statements=True,  # no Docstrings
            remove_annotations=True,  # not used runtime anyways
            hoist_literals=True,  # remove redundant strings
            rename_locals=True,  # short names save memory
            preserve_locals=["stubber"],  # names to keep
            rename_globals=True,  # short names save memory
            # keep these globals to allow testing/mocking to work agains the minified version
            preserve_globals=[
                "main",
                "Stubber",
                "read_path",
                "get_root",
                "_info",
                "os",
                "sys",
                "__version__",
            ],
            # remove_pass=True,  # no dead code
            # convert_posargs_to_args=True, # Does not save any space
        )
    log.debug(f"Original length : {len(content)}")
    log.debug(f"Minified length : {len(source)}")
    log.debug(f"Reduced by      : {len(content)-len(source)} ")
    return source


def minify(
    source: Union[str, Path],
    target: Union[str, Path],
    keep_report: bool = True,
    diff: bool = False,
    cross_compile: bool = False,
):

    source = Path(source)
    target = Path(target)
    # if target is a folder , then append the filename
    if target.exists() and target.is_dir():
        target = target / source.name
    try:
        with target.open("w+") as f:
            minified = minify_script(source_script=source, keep_report=keep_report, diff=diff)
            f.write(minified)
    except Exception as e:  # pragma: no cover
        log.exception(e)

    log.debug("Minified file written to :", target)
    if cross_compile:
        result = subprocess.run(["mpy-cross", "-O2", str(target)])
        if result.returncode == 0:
            log.debug("mpy-cross compiled to    :", target.with_suffix(".mpy"))
        return result.returncode
    else:
        return 0



##########################################################################################
# config
##########################################################################################
# pragma: no cover

from loguru import logger as log
from stubber.utils.config import CONFIG

from .cli import stubber_cli


@stubber_cli.command(name="show-config")
def cli_config():
    """
    Show the current configuration
    """
    try:
        # todo: sources[1] is depedent on the order the providers are added in
        log.info(f"config file                {CONFIG._provider._config_sources[1].filename}")  # type: ignore
    except Exception:
        pass
    log.info(f"CONFIG.repo_path           {CONFIG.repo_path}")
    log.info(f"CONFIG.mpy_path            {CONFIG.mpy_path}")
    log.info(f"CONFIG.mpy_lib_path        {CONFIG.mpy_lib_path}")

    log.info(f"CONFIG.stub_path           {CONFIG.stub_path}")
    log.info(f"CONFIG.publish_path        {CONFIG.publish_path}")
    log.info(f"CONFIG.template_path       {CONFIG.template_path}")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Create, Process, and Maintain stubs ✏️  for MicroPython"""



from stubber.commands.clone_cmd import cli_clone
from stubber.commands.config_cmd import cli_config
from stubber.commands.enrich_folder_cmd import cli_enrich_folder
from stubber.commands.get_core_cmd import cli_get_core
from stubber.commands.get_docstubs_cmd import cli_docstubs
from stubber.commands.get_frozen_cmd import cli_get_frozen
from stubber.commands.get_lobo_cmd import cli_get_lobo
from stubber.commands.merge_cmd import cli_merge_docstubs
from stubber.commands.minify_cmd import cli_minify
from stubber.commands.publish_cmd import cli_publish
from stubber.commands.stub_cmd import cli_stub
from stubber.commands.cli import stubber_cli
from stubber.commands.switch_cmd import cli_switch
from stubber.commands.upd_fallback_cmd import cli_update_fallback

##########################################################################################
# log = logging.getLogger("stubber")
#########################################################################################


##########################################################################################

if __name__ == "__main__":
    # add all commands to the CLI
    stubber_cli.add_command(cli_config)
    stubber_cli.add_command(cli_clone)
    stubber_cli.add_command(cli_switch)
    stubber_cli.add_command(cli_docstubs)
    stubber_cli.add_command(cli_get_core)
    stubber_cli.add_command(cli_get_frozen)
    stubber_cli.add_command(cli_get_lobo)
    stubber_cli.add_command(cli_stub)
    stubber_cli.add_command(cli_update_fallback)
    stubber_cli.add_command(cli_enrich_folder)
    stubber_cli.add_command(cli_minify)
    stubber_cli.add_command(cli_publish)
    stubber_cli.add_command(cli_merge_docstubs)
    stubber_cli()



"""
command line interface - main group
"""
import sys

import click
from loguru import logger as log
from stubber.utils.my_version import __version__


@click.group(chain=True)
@click.version_option(package_name="micropython-stubber", prog_name="micropython-stubber✏️ ")
@click.option(
    "-v",
    "--verbose",
    count=True,
    default=0,
    help="-v for DEBUG",
    is_eager=True,
)
@click.pass_context
def stubber_cli(ctx, verbose: int = 0):
    # ensure that ctx.obj exists and is a dict (in case `cli()` is called
    ctx.ensure_object(dict)
    # replace std log handler with a custom one capped on INFO level
    log.remove()
    level = {0: "INFO", 1: "DEBUG", 2: "TRACE"}.get(verbose, "INFO")
    if level == "INFO":
        format = "<green>{time:YY-MM-DD HH:mm:ss.SSS}</green> | <level>{level: <8}</level> | <cyan>{module: <18}</cyan> - <level>{message}</level>"
    else:
        format = "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | <level>{level: <8}</level> | <cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> - <level>{message}</level>"

    log.add(sys.stderr, level=level, backtrace=True, diagnose=True, colorize=True, format=format)
    log.info(f"micropython-stubber {__version__}")

    ctx.obj["loglevel"] = level
    ctx.obj["verbose"] = verbose



"""
Sort list of classess in parent-child order
note that this does not take multiple inheritance into account
ref : https://stackoverflow.com/questions/34964878/python-generate-a-dictionarytree-from-a-list-of-tuples/35049729#35049729
with modification 
"""
import re
from typing import List

from loguru import logger as log

__all__ = ["sort_classes"]
RE_CLASS = re.compile(r"class\s+(?P<class>\w+)(\((?P<parent>\w*)\))?")


def sort_classes(classes: List[str]):
    "sort a list of classes to respect the parent-child order"
    # Note this only takes single decedents in perspective
    # pass 1: create nodes dictionary
    nodes = {"": {"id": "root", "children": []}}

    for c in classes:
        m = RE_CLASS.match(c)
        if m:
            # parent_name = m.group("parent")
            class_name = m.group("class").strip()
            nodes[class_name] = {"id": class_name, "class": c, "children": []}

    # pass 2: create trees and parent-child relations
    forest = []
    forest.append(nodes[""])  #  add root node to connect
    for c in classes:
        m = RE_CLASS.match(c)
        if not m:
            continue
        if m.group("parent"):  # parent specified
            parent_name = m.group("parent").split(",")[0].strip()  # just use first parent
        else:
            parent_name = ""

        class_name = m.group("class").strip()
        node = nodes[class_name]

        # either make the node a new tree or link it to its parent
        if class_name == parent_name:
            # start a new tree in the forest
            forest.append(node)
        else:
            # add new_node as child to parent
            try:
                parent = nodes[parent_name]
            except KeyError:
                # Parent not defined in this module, add as child to root
                parent = nodes[""]
            if not "children" in parent:
                # ensure parent has a 'children' field
                parent["children"] = []
            children = parent["children"]
            children.append(node)  # type:ignore

    # step 3: simple function to print
    def list_node(node, sorted: List[str]):
        try:
            sorted.append(node["class"])
            log.trace(node["id"], node["class"])
        except KeyError:
            log.trace(node["id"])
        log.trace(node["id"])
        if node.get("children", False):
            for child in node["children"]:
                list_node(child, sorted)

    sorted: List[str] = []

    for node in forest:
        list_node(node, sorted)
    return sorted



"""Pre/Post Processing for createstubs.py"""
import subprocess
import sys
from pathlib import Path
from typing import List

from loguru import logger as log

from .stubmaker import generate_pyi_files

# # log = logging.getLogger(__name__)


def do_post_processing(stub_paths: List[Path], pyi: bool, black: bool):
    "Common post processing"
    for pth in stub_paths:
        if pyi:
            log.debug("Generate type hint files (pyi) in folder: {}".format(pth))
            generate_pyi_files(pth)
        if black:
            run_black(pth)


def run_black(path: Path):
    try:
        cmd = ["black", "."]

        if sys.version_info.major == 3 and sys.version_info.minor <= 7:
            # black on python 3.7 does not like some function defs
            # def sizeof(struct, layout_type=NATIVE, /) -> int:
            cmd += ["--fast"]
        # capture to suppress based on log level
        # result = subprocess.run(cmd, capture_output=log.level >= logging.INFO, check=True, shell=False, cwd=path)
        result = subprocess.run(cmd, capture_output=True, check=True, shell=False, cwd=path)
        if result.returncode != 0:  # pragma: no cover
            raise Exception(result.stderr.decode("utf-8"))
    except subprocess.SubprocessError:  # pragma: no cover
        log.error("some of the files are not in a proper format")



##########################################################################################
# clone
##########################################################################################
import os
from pathlib import Path
from typing import List, Tuple, Union

import click
import stubber.basicgit as git
from loguru import logger as log
from stubber.utils.config import CONFIG

from .cli import stubber_cli

##########################################################################################
# log = logging.getLogger("stubber")
#########################################################################################


@stubber_cli.command(name="clone")
@click.option("--path", "-p", default=CONFIG.repo_path.as_posix(), type=click.Path(file_okay=False, dir_okay=True))
@click.option("--add-stubs/--no-stubs", "stubs", default=False, is_flag=True, help="Also clone the micropython-stubs repo")
def cli_clone(path: Union[str, Path], stubs: bool = False):
    """
    Clone/fetch the micropython repos locally.

    The local repos are used to generate frozen-stubs and doc-stubs.
    """
    dest_path = Path(path)
    if not dest_path.exists():
        os.mkdir(dest_path)
    # repos are relative to provided path
    if dest_path != CONFIG.repo_path:
        mpy_path = dest_path / "micropython"
        mpy_lib_path = dest_path / "micropython-lib"
        mpy_stubs_path = dest_path / "micropython-stubs"
    else:
        mpy_path = CONFIG.mpy_path
        mpy_lib_path = CONFIG.mpy_lib_path

        mpy_stubs_path = CONFIG.stub_path.parent

    repos :List[Tuple[Path,str,str]]= [
        (mpy_path, "https://github.com/micropython/micropython.git", "master"),
        (mpy_lib_path, "https://github.com/micropython/micropython-lib.git", "master"),
    ]
    if stubs:
        repos.append((mpy_stubs_path, "https://github.com/josverl/micropython-stubs.git", "main"))

    for _path, remote, branch in repos:
        log.info(f"Cloning {remote} branch {branch} to {_path}")
        if not (_path / ".git").exists():
            log.debug(f"Cloning {_path.name}...")
            git.clone(remote_repo=remote, path=_path)
        else:
            log.debug(f"{_path.name} already exists, fetching...")
            git.fetch(
                _path,
            )
            git.pull(_path, branch=branch)  # DEFAULT

    log.info(f"{mpy_path} {git.get_tag(mpy_path)}")
    log.info(f"{mpy_lib_path} {git.get_tag(mpy_lib_path)}")
    # click.echo(f"{mpy_stubs_path} {git.get_tag(mpy_stubs_path)}")



"""
Classes and functions copied & adapted from micropypythons makemanifest.py to ensure that the manifest.py files can be processed
"""
import os
import shutil
from pathlib import Path
from typing import List, Union

from loguru import logger as log

# # log = logging.getLogger(__name__)
# log.setLevel(level=logging.DEBUG)

path_vars = {"MPY_DIR": "", "MPY_LIB_DIR": "", "PORT_DIR": "", "BOARD_DIR": ""}

# do not change class name
class FreezeError(Exception):
    pass


# do not change class name
class IncludeOptions:  # pragma: no cover
    def __init__(self, **kwargs):
        self._kwargs = kwargs
        self._defaults = {}

    def defaults(self, **kwargs):
        self._defaults = kwargs

    def __getattr__(self, name):
        return self._kwargs.get(name, self._defaults.get(name, None))


# do not change method name
# freeze_as_mpy is only used by the unix port.
def freeze_as_mpy(path, script=None, opt=0):
    log.trace(" - freeze_as_mpy({},script={},opt={})".format(path, script, opt))
    freeze(path, script, opt)


# do not change method name
def freeze_as_str(path):  # pragma: no cover
    log.trace(" - freeze_as_str({})".format(path))
    freeze(path)


# do not change method name
def freeze_mpy(path, script=None, opt=0):  # pragma: no cover
    """Freeze the input (see above), which must be .mpy files that are
    frozen directly.
    """
    log.trace(" - freeze_as_mpy({})".format(path))
    freeze(path, script)


# function used commonly in manifest.py to freeze a set of scripts
# pylint: disable=unused-argument, unused-variable
# do not change method name
def freeze(path, script=None, opt=0):  # pragma: no cover
    """Freeze the input, automatically determining its type.  A .py script
    will be compiled to a .mpy first then frozen, and a .mpy file will be
    frozen directly.

    `path` must be a directory, which is the base directory to search for
    files from.  When importing the resulting frozen modules, the name of
    the module will start after `path`, ie `path` is excluded from the
    module name.

    If `path` is relative, it is resolved to the current manifest.py.
    Use $(MPY_DIR), $(MPY_LIB_DIR), $(PORT_DIR), $(BOARD_DIR) if you need
    to access specific paths.

    If `script` is None all files in `path` will be frozen.

    If `script` is an iterable then freeze() is called on all items of the
    iterable (with the same `path` and `opt` passed through).

    If `script` is a string then it specifies the filename to freeze, and
    can include extra directories before the file.  The file will be
    searched for in `path`.

    `opt` is the optimisation level to pass to mpy-cross when compiling .py
    to .mpy. (ignored in this implementation)
    """
    log.trace(" - freeze(({},script={},opt={})".format(path, script, opt))
    path = convert_path(path)
    if script is None:
        # folder of scripts.
        # for s in os.listdir(path):
        #     freeze_internal(path, s)

        for dirpath, dirnames, filenames in os.walk(path, followlinks=True):  # type: ignore
            for script in filenames:
                # can recurse folder, so add relative path to script.
                freeze_internal(path, (dirpath + "/" + script)[len(path) + 1 :], opt)
    elif not isinstance(script, str):
        # several specific scripts.
        for script in script:
            freeze_internal(path, script)
    else:
        # on specific script, may include a path: 'umqtt/simple.py'
        freeze_internal(path, script)


# do not change method name
def include(manifest: Union[str,List[str]], **kwargs):
    """
    Include another manifest.

    The manifest argument can be a string (filename) or an iterable of
    strings.

    Relative paths are resolved with respect to the current manifest file.

    Optional kwargs can be provided which will be available to the
    included script via the `options` variable.

    e.g. include("path.py", extra_features=True)

    in path.py:
        options.defaults(standard_features=True)

        # freeze minimal modules.
        if options.standard_features:
            # freeze standard modules.
        if options.extra_features:
            # freeze extra modules.
    """
    if not isinstance(manifest, Path) and not isinstance(manifest, str): # type: ignore
        for m in manifest:
            include(m)
    else:
        manifest = convert_path(manifest)
        with open(manifest) as f:
            # Make paths relative to this manifest file while processing it.
            # Applies to includes and input files.
            prev_cwd = os.getcwd()
            os.chdir(os.path.dirname(manifest))
            try:
                # exec(f.read())  # pylint: disable=exec-used
                exec(f.read(), globals(), {"options": IncludeOptions(**kwargs)})  # pylint: disable=exec-used
            except OSError:
                log.warning("Could not process manifest: {}".format(manifest))
            os.chdir(prev_cwd)


# path_vars MUST be set externally
def convert_path(path: str) -> str:
    "Perform variable substitution in path"
    for name, value in path_vars.items():
        path = path.replace("$({})".format(name), value)
    # Convert to absolute path (so that future operations don't rely on
    # still being chdir 'ed).
    return os.path.abspath(path)


# -------------------------------------------------------------------------
# plug in the microsoft-stubber functionality
#

# stubdir MUST be set externally
stub_dir: str = ""
from pathlib import Path


# called by freeze.
# do not change method name
def freeze_internal(path: str, script: str, opt=None):
    """
    Copy the to-be-frozen module to the destination folder to be stubbed.

    Parameters:
    path (str)  : the source path
    script (str): the source script to be frozen
    opt (Any): freeze option (ignored)
    """

    log.trace(f" - freeze_internal({path},{script})")

    _path = Path(convert_path(path))
    source_path = _path / script
    dest_path = Path(stub_dir) / script

    if not _path.is_dir():  # pragma: no cover
        raise FreezeError("freeze source path should be a directory")
    if not stub_dir or stub_dir == "":  # pragma: no cover
        raise FreezeError("Stub folder not set")

    log.debug(f"freeze_internal : {script:<30} to {stub_dir}")
    # ensure folder, including possible path prefix for script
    os.makedirs(dest_path.parent, exist_ok=True)
    # copy file
    try:
        shutil.copy2(source_path, dest_path)
    except (FileNotFoundError) as e:  # pragma: no cover
        log.warning(f"File {path}/{script} not found")
    except (OSError, FileNotFoundError) as e:  # pragma: no cover
        log.exception(e)





import csv
import pkgutil
import tempfile
from collections import defaultdict
from pathlib import Path

import stubber.basicgit as git
from loguru import logger as log

# # log = logging.getLogger(__name__)


def switch(tag: str, *, mpy_path: Path, mpy_lib_path: Path):
    """
    Switch to a specific version of the micropython repos.

    Specify the version with --tag or --version to specify the version tag
    of the MicroPython repo.
    The Micropython-lib repo will be checked out to a commit that corresponds
    in time to that version tag, in order to allow non-current versions to be
    stubbed correctly.

    The repros must be cloned already
    """
    # fetch then switch
    git.fetch(mpy_path)
    git.fetch(mpy_lib_path)

    if not tag or tag in ["master", ""]:
        tag = "latest"
    if tag == "latest":
        git.switch_branch(repo=mpy_path, branch="master")
    else:
        git.checkout_tag(repo=mpy_path, tag=tag)
    match_lib_with_mpy(version_tag=tag, lib_path=mpy_lib_path)


def read_micropython_lib_commits(filename="data/micropython_tags.csv"):
    """
    Read a csv with the micropython version and matchin micropython-lib commit-hashes
    these can be used to make sure that the correct micropython-lib version is checked out.

    filename is relative to the 'stubber' package

    TODO: it would be nice if micropython-lib had matching commit-tags

        git for-each-ref --sort=creatordate --format '%(refname) %(creatordate)' refs/tags
    """
    data = pkgutil.get_data("stubber", filename)
    if not data:
        raise Exception(f"Resource {filename} not found")
    version_commit = defaultdict()  # lgtm [py/multiple-definition]
    with tempfile.NamedTemporaryFile(prefix="tags", suffix=".csv", mode="w+t") as ntf:
        ntf.file.write(data.decode(encoding="utf8"))
        ntf.file.seek(0)
        # read the csv file using DictReader
        reader = csv.DictReader(ntf.file, skipinitialspace=True)  # dialect="excel",
        rows = list(reader)
        # create a dict version --> commit_hash
        version_commit = {row["version"].split("/")[-1]: row["lib_commit_hash"] for row in rows if row["version"].startswith("refs/tags/")}
    # add default
    version_commit = defaultdict(lambda: "master", version_commit)
    return version_commit


def match_lib_with_mpy(version_tag: str, lib_path: Path):
    micropython_lib_commits = read_micropython_lib_commits()
    # Make sure that the correct micropython-lib release is checked out
    log.info(f"Matching repo's:  Micropython {version_tag} needs micropython-lib:{micropython_lib_commits[version_tag]}")
    return git.checkout_commit(micropython_lib_commits[version_tag], lib_path)



from typing import Optional

from packaging.version import Version, parse
from pypi_simple import PyPISimple


def get_pypy_versions( package_name:str, base : Optional[Version] = None, production:bool = True):
    """Get all versions of a package from a PyPI endpoint."""
    package_info = None
    if production:
        endpoint = "https://pypi.org/simple/"
    else:
        endpoint = "https://test.pypi.org/simple/"

    with PyPISimple(endpoint = endpoint) as client:
        package_info = client.get_project_page(project = package_name)

    if not package_info:
        return []

    versions = [parse(pkg.version) for pkg in package_info.packages if pkg.package_type == 'wheel' and pkg.version]
    # print(versions)

    if base:
        # if base provided then filter
        versions = [v for v in versions if v.base_version == base.base_version]

    return sorted(versions)



# pragma: no cover
"""
generate the list of modules that should be attempted to stub
for this : 
- combine the modules from the different texts files
- split the lines into individual module names
- combine them in one set
- remove the ones than cannot be stubbed
- remove test modules, ending in `_test`
- write updates to:
    - board/modulelist.txt
    - board/createsubs.py 

- TODO: remove the frozen modules from this list
- TODO: bump patch number if there are actual changes

"""

from pathlib import Path
from typing import Optional, Set, List, Union
from loguru import logger as log


def read_modules(path: Optional[Path] = None) -> Set:
    """
    read text files with modules per firmware.
    each contains the output of help("modules")
    - lines starting with # are comments.
    - split the other lines at whitespace seperator,
    - and add each module to a set
    """
    path = Path(path or "./data")
    assert path
    all_modules = set()
    for file in path.glob("*.txt"):
        log.debug("processing:", file)
        with file.open("r") as f:
            line = f.readline()
            while line:
                if len(line) > 1 and line[0] != "#":

                    file_mods = line.split()
                    log.trace(line[0:-1])
                    log.trace(set(file_mods))
                    # remove modules ending in _test
                    file_mods = [m for m in file_mods if not m.endswith("_test")]
                    all_modules = set(all_modules | set(file_mods))
                # next
                line = f.readline()

    log.trace(">" * 40)

    return all_modules


def wrapped(modules: Union[Set, List]) -> str:
    "wrap code line at spaces"
    long_line = str(modules)
    _wrapped = "        self.modules = "
    IDENT = len(_wrapped)
    MAX_WIDTH = 135

    # find seperator
    while len(long_line) > MAX_WIDTH:
        p1 = long_line.find("', ", MAX_WIDTH)
        # drop space
        p1 += 3
        short = long_line[0 : p1 - 1]
        _wrapped += short + "\n" + " " * IDENT
        long_line = long_line[p1 - 1 :]
    _wrapped += long_line
    return _wrapped


def main():
    """
    helper script
    generate a few lines of code with all modules to be stubbed by createstubs
    """
    #######################################################################
    # the exceptions
    #######################################################################
    mods_problematic = set(
        [
            "upysh",
            "webrepl_setup",
            "http_client",
            "http_client_ssl",
            "http_server",
            "http_server_ssl",
        ]
    )
    mods_excluded = set(
        [
            "__main__",
            "_main",
            "_boot",
            "webrepl",
            "_webrepl",
            "port_diag",
            "example_sub_led",
            "example_pub_button",
            "upip",
            "upip_utarfile",
            "upysh",
            # DOCSTUB_SKIP = [
            "uasyncio",  # can create better stubs from frozen python modules.
            "builtins",  # conflicts with static type checking , has very little information anyway
            "re",  # regex is too complex
            "collections",
            "io",
            "uio",
        ]
    )

    all_modules = read_modules()
    modules_to_stub = sorted(all_modules - set(mods_excluded | mods_problematic))

    # remove pycom MQTT* from defaults
    modules_to_stub = sorted({m for m in modules_to_stub if not m.startswith("MQTT")})

    log.trace("modules to stub :", len(modules_to_stub))
    log.trace(wrapped(modules_to_stub))

    # update modules.txt
    modules_txt = Path("board/modulelist.txt")
    if modules_txt.exists():
        with open(modules_txt) as f:
            lines = f.readlines()
            # only keep comment lines
            lines = [l for l in lines if l[0] == "#"]
    else:
        lines = ["# list of modules to stub."]
    lines = lines + [m + "\n" for m in modules_to_stub]

    with open(modules_txt, "w") as f:
        f.writelines(lines)

    # update createstubs.py
    createstubs = Path("board/createstubs.py")
    if createstubs.exists():
        with open(createstubs) as f:
            lines = f.readlines()

    l_start = lines.index("    stubber.modules = [\n")
    assert l_start
    l_end = lines.index("    ]  # spell-checker: enable\n", l_start)
    assert l_end

    # Plug in the new list of modules
    lines = lines[: l_start + 1] + [f'        "{m}",\n' for m in modules_to_stub] + lines[l_end:]

    with open(createstubs, "w") as f:
        f.writelines(lines)


if __name__ == "__main__":
    main()



# type: ignore
""" Work in Progress  

build test and  % report 
Will need to be updated after new_output has been implemented.

"""

# ref: https://regex101.com/codegen?language=python
# https://regex101.com/r/D5ddB2/1


import json
from pathlib import Path
from typing import List, Tuple

import stubber.utils as utils
from loguru import logger as log


def process(folder: Path, pattern: str):
    # read all .json files in the folder
    report = []
    for file in folder.glob(pattern):
        if file.name == "report.json":
            continue  # just skip the report to avoid errors
        with open(file, errors="ignore", encoding="utf8") as fp:
            # read docstrings from json ( unsafe , assumes much )
            module_name = file.stem
            docstrings: List[Tuple] = json.load(fp)
            for item in docstrings:
                # module, class, function/method , line, docstring
                if item[4] != []:

                    signature = str(item[3]).split("::")[-1].strip()
                    docstring = item[4]

                    r = utils._type_from_context(docstring=docstring, signature=signature, module=module_name)
                    report.append(
                        {
                            "signature": signature,
                            "docstring": docstring,
                            "docstring_len": len(" ".join(docstring).strip()),
                            "type": r["type"],
                            "confidence": r["confidence"],
                            "match": str(r["match"]),
                            "module": item[0],
                            "class": item[1],
                            "function/method": item[2],
                        }
                    )
                    # isGood = r["confidence"] >= 0.5 and r["confidence"] <= 0.8 and item[2] != ""
                    isBad = float(r["confidence"]) <= 0.5 and float(r["confidence"]) <= 0.8 and item[2] != ""
                    if isBad:
                        context = item[3] + ".".join((item[0], item[1], item[2]))
                        try:
                            log.debug(f"{context:40} {r['type']:<15} - {r['confidence']} {r['match'].groups('return')}")
                        except Exception:
                            log.debug(f"{context:40} {r['type']:<15} - {r['confidence']} ")

                        log.trace(r)
    if len(report) > 0:
        filename = folder / "report.json"
        with open(filename, mode="w", encoding="utf8") as fp:
            json.dump(report, fp, ensure_ascii=False, indent=4)


if __name__ == "__main__":
    process(Path("generated/micropython/1_16-nightly"), "*.json")
    # sample_authentication_with_api_key_credential()



""" 
Read the Micropython library documentation files and use them to build stubs that can be used for static typechecking 
using a custom-built parser to read and process the micropython RST files
- generates:
    - modules 
        - docstrings
    - function definitions 
        - function parameters based on documentation
        - docstrings
    - classes
        - docstrings
        - __init__ method
        - parameters based on documentation for class 
        - methods
            - parameters based on documentation for the method
            - docstrings

    - exceptions

- Tries to determine the return type by parsing the docstring.
    - Imperative verbs used in docstrings have a strong correlation to return -> None
    - recognizes documented Generators, Iterators, Callable
    - Coroutines are identified based tag "This is a Coroutine". Then if the return type was Foo, it will be transformed to : Coroutine[Foo]
    - a static Lookup list is used for a few methods/functions for which the return type cannot be determined from the docstring. 
    - add NoReturn to a few functions that never return ( stop / deepsleep / reset )
    - if no type can be detected the type `Any` is used

The generated stub files are formatted using `black` and checked for validity using `pyright`
Note: black on python 3.7 does not like some function defs 
`def sizeof(struct, layout_type=NATIVE, /) -> int:` 

- ordering of inter-dependent classes in the same module   

- Literals / constants
    - documentation contains repeated vars with the same indentation
    - Module level:
    .. code-block:: 

        .. data:: IPPROTO_UDP
                    IPPROTO_TCP

    - class level: 
    .. code-block:: 
    
        .. data:: Pin.IRQ_FALLING
                Pin.IRQ_RISING
                Pin.IRQ_LOW_LEVEL
                Pin.IRQ_HIGH_LEVEL

                Selects the IRQ trigger type.

    - literals documented using a wildcard are added as comments only 

- Add GLUE imports to allow specific modules to import specific others. 

- Repeats of definitions in the rst file for similar functions or literals
    - CONSTANTS ( module and Class level )
    - functions
    - methods

- Child/ Parent classes
    are added based on a (manual) lookup table CHILD_PARENT_CLASS

"""

import json
import os
import re
import subprocess
from pathlib import Path
from typing import List, Optional, Tuple

from loguru import logger as log

from . import utils
from .rst import (
    CHILD_PARENT_CLASS,
    DOCSTUB_SKIP,
    MODULE_GLUE,
    PARAM_FIXES,
    RST_DOC_FIXES,
    TYPING_IMPORT,
    U_MODULES,
    ClassSourceDict,
    FunctionSourceDict,
    ModuleSourceDict,
    return_type_from_context,
)
from .utils.config import CONFIG

# logging
# # log = logging.getLogger(__name__)

NEW_OUTPUT = True
#: self.gather_docs = True
SEPERATOR = "::"


def is_balanced(s: str) -> bool:
    """
    Check if a string has balanced parentheses
    """
    if s.count("(") != s.count(")"):
        return False
    if s.count("{") != s.count("}"):
        return False
    return True


class RSTReader:
    verbose = False
    gather_docs = False  # used only during Development
    target = ".py"  # py/pyi

    def __init__(self, v_tag="v1.xx"):
        self.line_no: int = 0  # current Linenumber used during parsing.
        self.filename = ""

        self.current_module = ""
        self.current_class = ""
        self.current_function = ""  # function & method

        # input buffer
        self.rst_text: List[str] = []
        self.max_line = 0

        # Output buffer
        self.output: List[str] = []
        self.output_dict: ModuleSourceDict = ModuleSourceDict("")
        self.output_dict.add_import(TYPING_IMPORT)

        self.source_tag = v_tag
        self.source_release = v_tag

        # development aids only
        self.return_info: List[Tuple] = []
        self.last_line = ""

    @property
    def line(self) -> str:
        "get the current line from input, also stores this as last_line to allow for inspection and dumping the json file"
        if self.line_no >= 0 and self.line_no <= self.max_line:
            self.last_line = self.rst_text[self.line_no]
        else:
            self.last_line = ""
        return self.last_line

    def extend_and_balance_line(self) -> str:
        """
        Append the current line + next line in order to try to balance the parentheses
        in order to do this the rst_test array is changed by the function
        and max_line is adjusted
        """
        append = 0
        newline = self.rst_text[self.line_no]
        while not is_balanced(newline) and self.line_no >= 0 and (self.line_no + append + 1) <= self.max_line:
            append += 1
            # concat the lines
            newline += self.rst_text[self.line_no + append]
        # only update line if things balanced out correctly
        if is_balanced(newline):
            self.rst_text[self.line_no] = newline
            for _ in range(append):
                self.rst_text.pop(self.line_no + 1)
                self.max_line -= 1
        # reprocess line
        return self.line

    @property
    def module_names(self) -> List[str]:
        "list of possible module names [uname , name] (longest first)"
        namelist: List[str] = []
        if self.current_module == "":
            return namelist
        # deal with module names "esp and esp.socket"
        if "." in self.current_module:
            names = [self.current_module, self.current_module.split(".")[0]]
        else:
            names = [self.current_module]
        # process
        for c_mod in names:
            if self.current_module[0] != "u":
                namelist += [f"u{c_mod}", c_mod]
            else:
                namelist += [c_mod, c_mod[1:]]
        return namelist

    def strip_prefixes(self, name: str, strip_mod: bool = True, strip_class: bool = False):
        "Remove the modulename. and or the classname. from the begining of a name"
        if strip_mod:
            prefixes = self.module_names
        else:
            prefixes = []
        if strip_class and self.current_class != "":
            prefixes += [self.current_class]
        for prefix in prefixes:
            if len(prefix) > 1 and prefix + "." in name:
                name = name.replace(prefix + ".", "")
        return name

    def leave_class(self):
        if self.current_class != "":
            self.current_class = ""

    def read_file(self, filename: Path):
        log.trace(f"Reading : {filename}")
        # ignore Unicode decoding issues
        with open(filename, errors="ignore", encoding="utf8") as file:
            self.rst_text = file.readlines()

        # Replace incorrect defenitions in .rst files with better ones
        for FIX in RST_DOC_FIXES:
            self.rst_text = [line.replace(FIX[0], FIX[1]) for line in self.rst_text]

        self.filename = filename.as_posix()  # use fwd slashes in origin
        self.max_line = len(self.rst_text) - 1
        self.current_module = filename.stem  # just to be sure

    def prepare_output(self):
        "clean up some trailing spaces and commas"
        if NEW_OUTPUT:
            lines = str(self.output_dict).splitlines(keepends=True)
            self.output = lines
        for i in range(0, len(self.output)):
            for name in ("self", "cls"):
                if f"({name}, ) ->" in self.output[i]:
                    self.output[i] = self.output[i].replace(f"({name}, ) ->", f"({name}) ->")

    def write_file(self, filename: Path) -> bool:
        self.prepare_output()
        try:
            log.debug(f" - Writing to: {filename}")
            with open(filename, mode="w", encoding="utf8") as file:
                file.writelines(self.output)
        except OSError as e:
            log.error(e)
            return False
        if self.gather_docs:
            log.debug(f" - Writing to: {filename.with_suffix('json')}")
            with open(filename.with_suffix(".json"), mode="w", encoding="utf8") as file:
                json.dump(self.return_info, file, ensure_ascii=False, indent=4)
            self.return_info = []

        return True

    def at_anchor(self) -> bool:
        "Stop at anchor ( however .. note: should be added)"
        _l = self.rst_text[self.line_no].lstrip()
        return _l.startswith("..") and not _l.startswith(".. note:")

    def at_heading(self) -> bool:
        "stop at heading"
        _l = self.rst_text[min(self.line_no + 1, self.max_line - 1)]
        # Heading  ---, ==, ~~~
        return _l.startswith("--") or _l.startswith("==") or _l.startswith("~~")

    def parse_docstring(self) -> List[str]:
        """Read a textblock that will be used as a docstring, or used to process a toc tree
        The textblock is terminated at the following RST line structures/tags
            .. <anchor>
            -- Heading
            == Heading
            ~~ Heading

        The blank lines at the start and end are removed to limit the space the docstring takes up.
        """
        if self.line_no >= len(self.rst_text):
            raise IndexError

        block: List[str] = []
        self.line_no += 1  # advance over current line
        try:
            while (
                self.line_no < len(self.rst_text)
                and not self.at_anchor()  # stop at next anchor ( however .. note: should be added)
                and not self.at_heading()  # stop at next heading
            ):
                line = self.rst_text[self.line_no]
                block.append(line.rstrip())
                self.line_no += 1  # advance line
        except IndexError:
            pass
        # if a Quoted Literal Block , then remove the first character of each line
        # https://docutils.sourceforge.io/docs/ref/rst/restructuredtext.html#quoted-literal-blocks
        if len(block) > 0 and len(block[0]) > 0 and block[0][0] != " ":
            q_char = block[0][0]
            if all([l.startswith(q_char) for l in block]):
                # all lines start with the same character, so skip that character
                block = [l[1:] for l in block]

        # remove empty lines at beginning/end of block
        if len(block) and len(block[0]) == 0:
            block = block[1:]
        if len(block) and len(block[-1]) == 0:
            block = block[:-1]

        # prettify
        if len(block):
            # Clean up Synopsis
            if ":synopsis:" in block[0]:
                block[0] = re.sub(
                    r"\s+:synopsis:\s+(?P<syn>[\w|\s]*)",
                    r"\g<syn>",
                    block[0],
                )
        # add clickable hyperlinks to CPython docpages
        for i in range(0, len(block)):

            # hyperlink to Cpython doc pages
            # https://regex101.com/r/5RN8rj/1
            # Optionally link to python 3.4 / 3.5 documentation
            _l = re.sub(
                r"(\s*\|see_cpython_module\|\s+:mod:`python:(?P<mod>[\w|\s]*)`)[.]?",
                r"\g<1> https://docs.python.org/3/library/\g<mod>.html .",
                block[i],
            )
            # RST hyperlink format is not clickable in v
            # https://regex101.com/r/5RN8rj/1
            _l = re.sub(
                r"(.*)(?P<url><https://docs\.python\.org/.*>)(`_)",
                r"\g<1>`\g<url>",
                _l,
            )
            # Clean up note
            _l = _l.replace(".. note:: ", "``Note:`` ")
            # clean up unsupported escape sequences in rst
            _l = _l.replace(r"\ ", " ")
            _l = _l.replace(r"\*", "*")
            block[i] = _l
        return block

    def fix_parameters(self, params: str):
        "change documentation parameter notation to a supported format that works for linting"
        params = params.strip()
        if not params.endswith(")"):
            # remove all after the closing bracket
            params = params[0 : params.rfind(")") + 1]

        ## Deal with SQUARE brackets first ( Documentation meaning := [optional])

        # multiple optionals
        # # .. method:: Servo.angle([angle, time=0])
        if "[angle, time=0]" in params:
            params = params.replace("[angle, time=0]", "[angle], time=0")
        # .. method:: Servo.speed([speed, time=0])
        elif "[speed, time=0]" in params:
            params = params.replace("[speed, time=0]", "[speed], time=0")

        # change [x] --> x:Optional[Any]
        params = params.replace("[", "")
        params = params.replace("]]", "")  # Q&D Hack-complex nesting

        # Handle Optional arguments
        # Optional step 1: [x] --> x: Optional[Any]=None
        params = params.replace("]", ": Optional[Any]=None")
        # Optional step 2: x: Optional[Any]=None='abc' --> x: Optional[Any]='abc'
        params = re.sub(r": Optional\[Any\]=None\s*=", r": Optional[Any]=", params)
        # Optional step 3: fix ...
        params = re.sub(r"\.\.\.: Optional\[Any\]=None", r"...: Optional[Any]", params)

        #
        # DOC: DocUpdate ? deal with overloads for Flash and Partition .readblock/writeblocks
        params = params.replace("block_num, buf, offset", "block_num, buf, offset: Optional[int]=0")

        # Remove modulename. and Classname. from class constant
        params = self.strip_prefixes(params, strip_mod=True, strip_class=True)

        for fix in PARAM_FIXES:
            if fix[0] in params:
                params = params.replace(fix[0], fix[1])

        # formatting
        # fixme: ... not allowed in .py
        if self.target == ".py":
            params = params.replace("*, ...", "*args")
            params = params.replace("...", "*args")

        return params

    def create_update_class(self, name: str, params: str, docstr: List[str]):
        # a bit of a hack: assume no classes in classes  or functions in function
        self.leave_class()
        # DOC: Add / update information to existing class definition
        full_name = self.output_dict.find(f"class {name}")
        if full_name:
            log.warning(f"TODO: UPDATE EXISTING CLASS : {name}")
            class_def = self.output_dict[full_name]
        else:
            # DOC: add the parent class(es) to the formal documentation
            parent = ""
            if name in CHILD_PARENT_CLASS.keys():
                parent = CHILD_PARENT_CLASS[name]
            if parent == "" and (name.endswith("Error") or name.endswith("Exception")):
                parent = "Exception"
            class_def = ClassSourceDict(
                f"class {name}({parent}):",
                docstr=docstr,
            )
        if len(params) > 0:
            method = FunctionSourceDict(
                name="__init__",
                indent=class_def.indent + 4,
                definition=[f"def __init__(self, {params} -> None:"],
                docstr=[],  # todo: check if twice is needed
            )
            class_def += method
        # Append class to output
        self.output_dict += class_def
        self.current_class = name

    def parse_toc(self):
        "process table of content with additional rst files, and add / include them in the current module"
        log.trace(f"# {self.line.rstrip()}")
        self.line_no += 1  # skip one line
        toctree = self.parse_docstring()
        # cleanup toctree
        toctree = [x.strip() for x in toctree if f"{self.current_module}." in x]
        # Now parse all files mentioned in the toc
        for file in toctree:
            #
            file_path = CONFIG.mpy_path / "docs" / "library" / file.strip()
            self.read_file(file_path)
            self.parse()
        # reset this file to done
        self.rst_text = []
        self.line_no = 1

    def parse_module(self):
        "parse a module tag and set the module's docstring"
        log.trace(f"# {self.line.rstrip()}")
        module_name = self.line.split(SEPERATOR)[-1].strip()

        self.current_module = module_name
        self.current_function = self.current_class = ""
        # get module docstring
        docstr = self.parse_docstring()

        if len(docstr) > 0:
            # Add link to online documentation
            # https://docs.micropython.org/en/v1.17/library/array.html
            if "nightly" in self.source_tag:
                version = "latest"
            else:
                version = self.source_tag.replace("_", ".")
            docstr[0] = docstr[0] + f". See: https://docs.micropython.org/en/{version}/library/{module_name}.html"

        self.output_dict.name = module_name
        self.output_dict.add_comment(f"# source version: {self.source_tag}")
        self.output_dict.add_comment(f"# origin module:: {self.filename}")
        self.output_dict.add_docstr(docstr)
        # Add additional imports to allow one module te refer to another
        if module_name in MODULE_GLUE.keys():
            self.output_dict.add_import(MODULE_GLUE[module_name])

    def parse_current_module(self):
        log.trace(f"# {self.line.rstrip()}")
        module_name = self.line.split(SEPERATOR)[-1].strip()
        mod_comment = f"# + module: {self.current_module}.rst"
        self.current_module = module_name
        self.current_function = self.current_class = ""
        log.debug(mod_comment)
        if NEW_OUTPUT:  # new output
            self.output_dict.name = module_name
            self.output_dict.add_comment(mod_comment)

        self.line_no += 1  # advance as we did not read any docstring
        # todo: read first block and do something with it ( for a submodule)

    def parse_function(self):
        log.trace(f"# {self.line.rstrip()}")
        # this_function = self.line.split(SEPERATOR)[-1].strip()
        # name = this_function

        # Get one or more names
        function_names = self.parse_names(oneliner=False)
        docstr = self.parse_docstring()

        for this_function in function_names:
            # Parse return type from docstring
            ret_type = return_type_from_context(docstring=docstr, signature=this_function, module=self.current_module)

            ...
            # defaults
            name = params = ""
            try:
                name, params = this_function.split("(", maxsplit=1)
            except ValueError:
                log.error(this_function)
            self.current_function = name
            if name in ("classmethod", "staticmethod"):
                # skip the classmethod and static method functions
                # no use to create stubs for these
                return

            # ussl documentation uses a ssl.foobar prefix
            for mod in self.module_names:
                if name.startswith(f"{mod}."):
                    # remove module name from the start of the function name
                    name = name[len(f"{mod}.") :]
            # fixup parameters
            params = self.fix_parameters(params)
            # assume no functions in classes
            self.leave_class()
            # if function name is the same as the module
            # then this is probably documenting a class ()

            # if the function name matches the module name then threat this as a class.
            if name in self.module_names:
                # 'Promote' function to class
                self.create_update_class(name, params, docstr)
            else:
                fn_def = FunctionSourceDict(
                    name=f"def {name}",
                    definition=[f"def {name}({params} -> {ret_type}:"],
                    docstr=docstr,
                )
                self.output_dict += fn_def

    def parse_class(self):
        log.trace(f"# {self.line.rstrip()}")
        this_class = self.line.split(SEPERATOR)[-1].strip()  # raw
        if "(" in this_class:
            name, params = this_class.split("(", 2)
        else:
            name = this_class
            params = ""
        name = self.strip_prefixes(name)
        self.current_class = name
        self.current_function = ""

        log.trace(f"# class:: {name} - {this_class}")
        # fixup parameters
        params = self.fix_parameters(params)
        docstr = self.parse_docstring()

        if any(":noindex:" in line for line in docstr):
            # if the class docstring contains ':noindex:' on any line then skip
            log.trace(f"# Skip :noindex: class {name}")
        else:
            # write a class header
            self.create_update_class(name, params, docstr)

    def get_rst_hint(self):
        "parse the '.. <rst hint>:: ' from the current line"
        m = re.search(r"\.\.\s?(\w+)\s?::\s?", self.line)
        if m:
            return m.group(1)
        else:
            return ""

    def parse_method(self):
        name = ""
        this_method = ""
        # params = ")"
        ## py:staticmethod  - py:classmethod - py:decorator
        # ref: https://sphinx-tutorial.readthedocs.io/cheatsheet/
        log.trace(f"# {self.line.rstrip()}")
        if not is_balanced(self.line):
            self.extend_and_balance_line()

        ## rst_hint is used to access the method decorator ( method, staticmethod, staticmethod ... )
        rst_hint = self.get_rst_hint()

        method_names = self.parse_names(oneliner=False)
        # filter out overloads with 'param=value' description as these can't be output (currently)
        method_names = [m for m in method_names if not "param=value" in m]

        docstr = self.parse_docstring()
        for this_method in method_names:
            try:
                name, params = this_method.split("(", 1)  # split methodname from params
            except ValueError:
                name = this_method
                params = ")"
            self.current_function = name
            # self.writeln(f"# method:: {name}")
            if "." in name:
                # todo deal with longer / deeper classes
                class_name = name.split(".")[0]
                # update current for out-of sequence method processing
                self.current_class = class_name
            else:
                # if nothing specified lets assume part of current class
                class_name = self.current_class
            name = name.split(".")[-1]  # Take only the last part from Pin.toggle

            # fixup optional [] parameters and other notations
            params = self.fix_parameters(params)

            # get or create the parent class
            full_name = self.output_dict.find(f"class {class_name}")
            if full_name:
                parent_class = self.output_dict[full_name]
            else:
                # not found, create and add new class to the output dict
                parent_class = ClassSourceDict(f"class {class_name}():")
                self.output_dict += parent_class

            # parse return type from docstring
            ret_type = return_type_from_context(docstring=docstr, signature=f"{class_name}.{name}", module=self.current_module)
            # methods have 4 flavours
            #   - __init__              (self,  <params>) -> None:
            #   - classmethod           (cls,   <params>) -> <ret_type>:
            #   - staticmethod          (       <params>) -> <ret_type>:
            #   - all other methods     (self,  <params>) -> <ret_type>:
            if name == "__init__":
                method = FunctionSourceDict(
                    name=f"def {name}",
                    indent=parent_class.indent + 4,
                    definition=[f"def __init__(self, {params} -> None:"],
                    docstr=docstr,
                )
                parent_class += method
            elif rst_hint == "classmethod":
                method = FunctionSourceDict(
                    decorators=["@classmethod"],
                    name=f"def {name}",
                    indent=parent_class.indent + 4,
                    definition=[f"def {name}(cls, {params} -> {ret_type}:"],
                    docstr=docstr,
                )
                parent_class += method

            elif rst_hint == "staticmethod":
                method = FunctionSourceDict(
                    decorators=["@staticmethod"],
                    name=f"def {name}",
                    indent=parent_class.indent + 4,
                    definition=[f"def {name}({params} -> {ret_type}:"],
                    docstr=docstr,
                )
                parent_class += method
            else:  # just plain method
                method = FunctionSourceDict(
                    name=f"def {name}",
                    indent=parent_class.indent + 4,
                    definition=[f"def {name}(self, {params} -> {ret_type}:"],
                    docstr=docstr,
                )
                parent_class += method

    def parse_exception(self):
        log.trace(f"# {self.line.rstrip()}")
        name = self.line.split(SEPERATOR)[1].strip()
        if name == "Exception":
            # no need to redefine Exception
            self.line_no += 1
            return
        # Take only the last part from module.ExceptionX
        if "." in name:
            name = name.split(".")[-1]
        except_1 = ClassSourceDict(name=f"class {name}(Exception) : ...", docstr=[], init="")
        self.output_dict += except_1
        # no docstream read (yet) , so need to advance to next line
        self.line_no += 1

    def parse_name(self, line: Optional[str] = None):
        "get the constant/function/class name from a line with an identifier"
        # this_const = self.line.split(SEPERATOR)[-1].strip()
        # name = this_const
        # name = this_const.split(".")[1]  # Take only the last part from Pin.toggle

        if line:
            return line.split(SEPERATOR)[-1].strip()
        else:
            return self.line.split(SEPERATOR)[-1].strip()

    def parse_names(self, oneliner: bool = True):
        """get a list of constant/function/class names from and following a line with an identifier
        advances the linecounter

        oneliner :  treat a line with commas as multiple names (used for constants)
        """
        names: List[str] = []
        if oneliner:
            names += self.parse_name().split(",")
        else:
            names += [self.parse_name()]

        m = re.search(r"..\s?\w+\s?::\s?", self.line)
        if not m:  # pragma: no cover
            raise KeyError
        col = m.end()
        counter = 1
        while (
            self.line_no + counter <= self.max_line
            and self.rst_text[self.line_no + counter].startswith(" " * col)
            and not self.rst_text[self.line_no + counter][col + 1].isspace()
        ):
            if self.verbose:  # pragma: no cover
                log.trace("Sequence detected")
            names.append(self.parse_name(self.rst_text[self.line_no + counter]))
            counter += 1
        # now advance the linecounter
        self.line_no += counter - 1
        # clean up before returning
        return [n.strip() for n in names if n.strip() != "etc."]

    def parse_data(self):
        log.trace(f"# {self.line.rstrip()}")
        # Get one or more names
        names = self.parse_names()

        # get module docstring
        docstr = self.parse_docstring()

        # deal with documentation wildcards
        for name in names:

            type = return_type_from_context(docstring=docstr, signature=name, module=self.current_module, literal=True)
            if type in ["None"]:  # None does not make sense
                type = "Any"  # perhaps default to Int ?
            name = self.strip_prefixes(name)
            self.output_dict.add_constant_smart(name=name, type=type, docstr=docstr)

    def parse(self):
        self.line_no = 0
        while self.line_no < len(self.rst_text):
            line = self.line
            rst_hint = self.get_rst_hint()
            #    self.writeln(">"+line)
            if rst_hint == "module":
                self.parse_module()
            elif rst_hint == "currentmodule":
                self.parse_current_module()
            elif rst_hint == "function":
                self.parse_function()

            elif rst_hint == "class":
                self.parse_class()
            elif rst_hint == "method" or rst_hint == "staticmethod" or rst_hint == "classmethod":
                self.parse_method()
            elif rst_hint == "exception":
                self.parse_exception()
            elif rst_hint == "data":
                self.parse_data()

            elif rst_hint == "toctree":
                self.parse_toc()
                # not this will be the end of this file processing.

            elif len(rst_hint) > 0:
                # something new / not yet parsed
                self.line_no += 1
                log.trace(f"# {line.rstrip()}")
            else:
                # NOTHING TO SEE HERE , MOVE ON
                self.line_no += 1
            #  TODO: rebuild this into the NEW_OUTPUT way of working
            #     if self.gather_docs and len(docstr) > 0:
            #         self.return_info.append(
            #             (
            #                 self.current_module,
            #                 self.current_class,
            #                 self.current_function,
            #                 self.last_line,
            #                 docstr,
            #             )
            #         )

        # now clean up
        self.prepare_output()


def generate_from_rst(
    rst_path: Path,
    dst_path: Path,
    v_tag: str,
    release: Optional[str] = None,
    pattern: str = "*.rst",
    verbose: bool = False,
    suffix=".py",
) -> int:
    if not dst_path.exists():
        dst_path.mkdir(parents=True)
    if not release:
        release = v_tag
    # skip
    #  - index,
    # - modules with a . in the stem :  module.xxx.rst is included in module.py

    files = [f for f in rst_path.glob(pattern) if f.stem != "index" and "." not in f.stem]

    # - excluded modules, ones that offer little advantage  or cause much problems
    files = [f for f in files if f.name not in DOCSTUB_SKIP]

    # simplify debugging
    # files = [f for f in files if f.name == "collections.rst"]

    # remove all .py/.pyi files in desination folder to avoid left-behinds
    for f in dst_path.rglob(pattern="*.py*"):
        try:
            os.remove(f)
        except (OSError, PermissionError):
            pass

    for file in files:
        reader = RSTReader(v_tag)
        reader.source_release = release
        reader.verbose = verbose
        log.debug(f"Reading: {file}")
        reader.read_file(file)
        reader.parse()
        if file.stem in U_MODULES:
            # create umod.py file and mod.py file
            reader.write_file((dst_path / ("u" + file.name)).with_suffix(suffix))
            reader.write_file((dst_path / file.name).with_suffix(suffix))
            # with open((dst_path / file.name).with_suffix(suffix), "w") as new_file:
            #     # new_file.write(f"from u{file.stem} import * # Type: Ignore\n")
            #     new_file.write(f"from . import u{file.stem} # Type: Ignore\n")
        else:
            # create mod.py file
            reader.write_file((dst_path / file.name).with_suffix(suffix))
        del reader

    # run autoflake to remove unused imports
    # needs to be run BEFORE black otherwise it does not recognize long import from`s.
    cmd = [
        "autoflake",
        "-r",
        "--in-place",
        dst_path.as_posix(),
        "-v",
        "-v",  # show some feedback
    ]
    # subprocess.run(cmd, capture_output=log.level >= logging.INFO)
    subprocess.run(cmd, capture_output=True)

    # Also generate a module manifest
    utils.make_manifest(
        folder=dst_path,
        family="micropython",
        version=utils.clean_version(v_tag),
        release=release,
        port="-",
        stubtype="documentation",
    )

    return len(files)



from pathlib import Path
from typing import Any, Dict, Optional

# from stubber.codemod.merge_docstub import MergeCommand
import stubber.codemod.merge_docstub as merge_docstub
from libcst.codemod import (CodemodContext, diff_code,
                            exec_transform_with_prettyprint)
from libcst.tool import _default_config
from loguru import logger as log

##########################################################################################
# # log = logging.getLogger(__name__)
# logging.basicConfig(level=logging.INFO)
#########################################################################################


def enrich_file(source_path: Path, docstub_path: Path, diff=False, write_back=False) -> Optional[str]:
    """\
    Enrich a firmware stubs using the doc-stubs in another folder.
    Both (.py or .pyi) files are supported.

    Parameters:
        source_path: the path to the firmware stub to enrich
        docstub_path: the path to the folder containg the doc-stubs
        diff: if True, return the diff between the original and the enriched source file
        write_back: if True, write the enriched source file back to the source_path

    Returns:
    - None or a string containing the diff between the original and the enriched source file
    """
    config: Dict[str, Any] = _default_config()
    context = CodemodContext()

    # find a matching doc-stub file in the docstub_path
    docstub_file = None
    for ext in [".py", ".pyi"]:
        candidates = list(docstub_path.rglob(source_path.stem + ext))
        if source_path.stem[0].lower() == "u":
            # also look for candidates without leading u
            candidates += list(docstub_path.rglob(source_path.stem[1:] + ext))
        else:
            # also look for candidates with leading u
            candidates += list(docstub_path.rglob("u" + source_path.stem + ext))

        for docstub_file in candidates:
            if docstub_file.exists():
                break
            else:
                docstub_file = None
    if docstub_file is None:
        raise FileNotFoundError(f"No doc-stub file found for {source_path}")

    log.debug(f"Merge {docstub_file} from {source_path}")
    # read source file
    oldcode = source_path.read_text()

    codemod_instance = merge_docstub.MergeCommand(context, stub_file=docstub_file)
    newcode = exec_transform_with_prettyprint(
        codemod_instance,
        oldcode,
        # include_generated=False,
        generated_code_marker=config["generated_code_marker"],
        # format_code=not args.no_format,
        formatter_args=config["formatter"],
        # python_version=args.python_version,
    )
    if newcode:
        if write_back:
            log.trace(f"Write back enriched file {source_path}")
            # write updated code to file
            source_path.write_text(newcode, encoding="utf-8")
        if diff:  # pragma: no cover
            return diff_code(oldcode, newcode, 5, filename=source_path.name)
        return newcode
    else:
        return None


def enrich_folder(source_folder: Path, docstub_path: Path, show_diff=False, write_back=False, require_docstub=False) -> int:
    """\
        Enrich a folder with containing firmware stubs using the doc-stubs in another folder.
        
        Returns the number of files enriched.
    """
    count = 0
    # list all the .py and .pyi files in the source folder
    source_files = sorted(list(source_folder.rglob("**/*.py")) + list(source_folder.rglob("**/*.pyi")))
    for source_file in source_files:
        try:
            diff = enrich_file(source_file, docstub_path, diff=True, write_back=write_back)
            if diff:
                count += 1
                if show_diff:
                    print(diff)
        except FileNotFoundError:
            # no docstub to enrich with
            if require_docstub:
                raise (FileNotFoundError(f"No doc-stub file found for {source_file}"))
    return count



from enum import Enum


class StubSource(str, Enum):
    FIRMWARE = "Firmware stubs"
    "stubs built by combining the firmware, frozen and core stubs"
    FROZEN = "Frozen stubs"
    "stubs of python modules that are frozen as part of the firmware image"
    CORE = "Core stubs"
    "stubs that allow (some) MicroPython code to be run by CPython"
    DOC = "Doc stubs"
    "stubs built by parsing the micropython RST documentation files"
    MERGED = "Merged stubs"
    "stubs built by merging the information from doc-stubs and firmware-stubs"


ALL_TYPES = ["combo", "doc", "core", "firmware"]
COMBO_STUBS = ALL_TYPES[0]
DOC_STUBS = ALL_TYPES[1]
CORE_STUBS = ALL_TYPES[2]
FIRMWARE_STUBS = ALL_TYPES[3]

# # TODO: Combine the type classes ?
# class StubTypes(str, Enum):
#     COMBO = "combo"
#     "stubs built by combining the firmware, doc and core stubs"
#     DOC = "doc"
#     "stubs built by parsing  the RST documentation files"
#     CORE = "core"
#     "stubs that allow (some) MicroPython code to be run by CPython"



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['src',
 'src.stubber',
 'src.stubber.codemod',
 'src.stubber.commands',
 'src.stubber.freeze',
 'src.stubber.publish',
 'src.stubber.rst',
 'src.stubber.tools',
 'src.stubber.utils']

package_data = \
{'': ['*'], 'src.stubber': ['data/*']}

install_requires = \
['PyGithub==1.55',
 'autoflake==1.4',
 'black==22.6.0',
 'click==8.0.3',
 'libcst==0.4.7',
 'loguru==0.6.0',
 'mpy-cross==1.19.1',
 'mypy==0.971',
 'packaging==21.3',
 'pypi-simple==0.10.0',
 'pysondb-v2==2.1.0',
 'python-minifier==2.6.0',
 'requests==2.27.0',
 'tabulate==0.8.10',
 'tomli-w==1.0.0',
 'typed-config==1.1.0']

extras_require = \
{':python_version < "3.11"': ['tomli==2.0.1'],
 ':python_version < "3.8"': ['importlib-metadata==1.0'],
 'tools': ['esptool==3.3.1', 'rshell==0.0.31', 'pyserial==3.5']}

entry_points = \
{'console_scripts': ['pyboard = stubber.tools.pyboard:main',
                     'stubber = stubber.stubber:stubber_cli']}

setup_kwargs = {
    'name': 'micropython-stubber',
    'version': '1.9.8',
    'description': 'Tooling to create and maintain stubs for MicroPython',
    'long_description': '# Boost MicroPython productivity in VSCode\n \n  [![pypi version](https://badgen.net/pypi/v/micropython-stubber)](https://pypi.org/project/micropython-stubber/)\n  [![python versions](https://badgen.net/pypi/python/micropython-stubber)](https://badgen.net/pypi/python/micropython-stubber)\n  [![Documentation Status](https://readthedocs.org/projects/micropython-stubber/badge/?version=latest)](https://micropython-stubber.readthedocs.io/en/latest/?badge=latest "Document build status badge")\n  [![Star on GitHub](https://img.shields.io/github/stars/josverl/micropython-stubber.svg?style=social)](https://github.com/josverl/micropython-stubber/stargazers)\n  [![All Contributors](https://img.shields.io/badge/all_contributors-19-green.svg?style=flat-square)](#Contributions)\n  <!-- break -->\n  [![Language grade: Python](https://img.shields.io/lgtm/grade/python/g/Josverl/micropython-stubber.svg?logo=lgtm&logoWidth=18)](https://lgtm.com/projects/g/Josverl/micropython-stubber/context:python)\n  [![pytest tests/common](https://github.com/Josverl/micropython-stubber/actions/workflows/pytest.yml/badge.svg)](https://github.com/Josverl/micropython-stubber/actions/workflows/pytest.yml)\n  [![codecov](https://codecov.io/gh/Josverl/micropython-stubber/branch/main/graph/badge.svg?token=WJFGMKBHOV)](https://codecov.io/gh/Josverl/micropython-stubber)\n  [![Black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/python/black "Black badge")\n  <!-- break -->\n  [![Open in VSCode](https://img.shields.io/static/v1?logo=visualstudiocode&label=&message=Open%20in%20Visual%20Studio%20Code&labelColor=2c2c32&color=007acc&logoColor=007acc\n)](https://open.vscode.dev/josverl/micropython-stubber)\n  \nThe intellisense and code linting that is so prevalent in modern editors, does not work out-of-the-gate for MicroPython projects.\nWhile the language is Python, the modules used are different from CPython , and also different ports have different modules and classes , or the same class with different parameters.\n\nWriting MicroPython code in a modern editor should not need to involve keeping a browser open to check for the exact parameters to read a sensor, light-up a led or send a network request.\n\nFortunately with some additional configuration and data, it is possible to make the editors understand your flavor of MicroPython, wether you use one of the pre-compiled firmwares, but also if you run a one-off custom firmware version.\n\n\n![demo][]]\n\nIn order to achieve this a few things are needed:\n1) Stub files for the native / enabled modules in the firmware using PEP 484 Type Hints\n2) Specific configuration of the VSCode Python extensions \n3) Specific configuration of Pylint [ Optional ]\n4) Suppression of warnings that collide with the MicroPython principals or code optimization.\n\nPlease review the documentation on [https://micropython-stubber.readthedocs.io]  \n\nWith that in place, VSCode will understand MicroPython for the most part, and help you to write code, and catch more errors before deploying it to your board. \n\nNote that the above is not limited to VSCode and pylint, but it happens to be the combination that I use. \n\nA lot of subs have already been generated and are shared on github or other means,  so it is quite likely that you can just grab a copy be be productive in a few minutes.\n\nFor now you will need to [configure this by hand](#manual-configuration), or use the [micropy cli` tool](#using-micropy-cli)\n\n1. The sister-repo [**MicroPython-stubs**][stubs-repo] contains [all stubs][all-stubs] I have collected with the help of others, and which can be used directly.\nThat repo also contains examples configuration files that can be easily adopted to your setup.\n\n2. A second repo [micropy-stubs repo][stubs-repo2] maintained by BradenM, also contains stubs, but in a structure only used and distributed by the [micropy-cli](#using-micropy-cli) tool.\nYou should use micropy-cli to consume stubs from that repo.\n\nThe (stretch) goal is to create a VSCode add-in to simplify the configuration, and allow easy switching between different firmwares and versions.\n\n\n## Install and basic usage\n\n``` sh\npip install micropython-stubber\n\n# go to your working folder \ncd my_stub_folder\nmkdir all-stubs\n\n# clone the micropython repo\'s and switch to a specific version \nstubber clone\nstubber switch --version v1.18\n\n# get the document stubs for the current version ( v1.18 )\nstubber get-docstubs\n\n# get the frozen stubs for the current version ( v1.18 )\nstubber get-frozen\n\n# get the core CPython compatibility stubs from PyPi \nstubber get-core\n\n# Update the fallback stubs\nstubber update-fallback\n\n#\nls all-stubs\ndir all-stubs\n```\n\n\n## Developing & testing \n\nThis is described in more detail in the [developing](docs/developing.md) and [testing](docs/testing.md)  documents in the docs folder.\n\n## Branch Main\nThe name of the default branch has been changed to `main`.\nIf you have cloned this repo before you main need to adjust the local repro to be aware of this, or create a fresh clone.\n\nTo update run the below command:  \n``` bash\ngit branch -m master main                    \ngit fetch origin\ngit branch -u origin/main main                      \ngit remote set-head origin -a\n```\n\nfor more info see [**Renaming a branch**](https://docs.github.com/en/repositories/configuring-branches-and-merges-in-your-repository/managing-branches-in-your-repository/renaming-a-branch#updating-a-local-clone-after-a-branch-name-changes)\n\n## Licensing \n\nMicroPython-Stubber is licensed under the MIT license, and all contributions should follow this [LICENSE](LICENSE).\n\n\n# Contributions\n<!-- spell-checker: disable -->\n\n<!-- ALL-CONTRIBUTORS-LIST:START - Do not remove or modify this section -->\n<!-- prettier-ignore-start -->\n<!-- markdownlint-disable -->\n<table>\n  <tr>\n    <td align="center"><a href="https://github.com/Josverl"><img src="https://avatars2.githubusercontent.com/u/981654?v=4?s=100" width="100px;" alt=""/><br /><sub><b>Jos Verlinde</b></sub></a><br /><a href="https://github.com/Josverl/micropython-stubber/commits?author=josverl" title="Code">💻</a> <a href="#research-josverl" title="Research">🔬</a> <a href="#ideas-josverl" title="Ideas, Planning, & Feedback">🤔</a> <a href="#content-josverl" title="Content">🖋</a> <a href="#stubs-josverl" title="MicroPython stubs">📚</a> <a href="#test-josverl" title="Test">✔</a></td>\n    <td align="center"><a href="https://thonny.org/"><img src="https://avatars1.githubusercontent.com/u/46202078?v=4?s=100" width="100px;" alt=""/><br /><sub><b>Thonny, Python IDE for beginners</b></sub></a><br /><a href="#ideas-thonny" title="Ideas, Planning, & Feedback">🤔</a> <a href="#research-thonny" title="Research">🔬</a></td>\n    <td align="center"><a href="https://micropython.org/"><img src="https://avatars1.githubusercontent.com/u/6298560?v=4?s=100" width="100px;" alt=""/><br /><sub><b>MicroPython</b></sub></a><br /><a href="#data-micropython" title="Data">🔣</a> <a href="#stubs-micropython" title="MicroPython stubs">📚</a></td>\n    <td align="center"><a href="https://github.com/loboris"><img src="https://avatars3.githubusercontent.com/u/6280349?v=4?s=100" width="100px;" alt=""/><br /><sub><b>Boris Lovosevic</b></sub></a><br /><a href="#data-loboris" title="Data">🔣</a> <a href="#stubs-loboris" title="MicroPython stubs">📚</a></td>\n    <td align="center"><a href="https://github.com/pfalcon"><img src="https://avatars3.githubusercontent.com/u/500451?v=4?s=100" width="100px;" alt=""/><br /><sub><b>Paul Sokolovsky</b></sub></a><br /><a href="#data-pfalcon" title="Data">🔣</a> <a href="#stubs-pfalcon" title="MicroPython stubs">📚</a></td>\n    <td align="center"><a href="https://github.com/pycopy"><img src="https://avatars0.githubusercontent.com/u/67273174?v=4?s=100" width="100px;" alt=""/><br /><sub><b>pycopy</b></sub></a><br /><a href="#data-pycopy" title="Data">🔣</a> <a href="#stubs-pycopy" title="MicroPython stubs">📚</a></td>\n    <td align="center"><a href="https://github.com/pycom"><img src="https://avatars2.githubusercontent.com/u/16415153?v=4?s=100" width="100px;" alt=""/><br /><sub><b>Pycom</b></sub></a><br /><a href="#infra-pycom" title="Infrastructure (Hosting, Build-Tools, etc)">🚇</a></td>\n  </tr>\n  <tr>\n    <td align="center"><a href="https://github.com/BradenM"><img src="https://avatars1.githubusercontent.com/u/5913808?v=4?s=100" width="100px;" alt=""/><br /><sub><b>Braden Mars</b></sub></a><br /><a href="https://github.com/Josverl/micropython-stubber/issues?q=author%3ABradenM" title="Bug reports">🐛</a> <a href="https://github.com/Josverl/micropython-stubber/commits?author=BradenM" title="Code">💻</a> <a href="#stubs-BradenM" title="MicroPython stubs">📚</a> <a href="#platform-BradenM" title="Packaging/porting to new platform">📦</a></td>\n    <td align="center"><a href="https://binary.com.au/"><img src="https://avatars2.githubusercontent.com/u/175909?v=4?s=100" width="100px;" alt=""/><br /><sub><b>James Manners</b></sub></a><br /><a href="https://github.com/Josverl/micropython-stubber/commits?author=jmannau" title="Code">💻</a> <a href="https://github.com/Josverl/micropython-stubber/issues?q=author%3Ajmannau" title="Bug reports">🐛</a></td>\n    <td align="center"><a href="http://patrickwalters.us/"><img src="https://avatars0.githubusercontent.com/u/4002194?v=4?s=100" width="100px;" alt=""/><br /><sub><b>Patrick</b></sub></a><br /><a href="https://github.com/Josverl/micropython-stubber/issues?q=author%3Aaskpatrickw" title="Bug reports">🐛</a> <a href="https://github.com/Josverl/micropython-stubber/commits?author=askpatrickw" title="Code">💻</a> <a href="#stubs-askpatrickw" title="MicroPython stubs">📚</a></td>\n    <td align="center"><a href="https://opencollective.com/pythonseverywhere"><img src="https://avatars3.githubusercontent.com/u/16009100?v=4?s=100" width="100px;" alt=""/><br /><sub><b>Paul m. p. P.</b></sub></a><br /><a href="#ideas-pmp-p" title="Ideas, Planning, & Feedback">🤔</a> <a href="#research-pmp-p" title="Research">🔬</a></td>\n    <td align="center"><a href="https://github.com/edreamleo"><img src="https://avatars0.githubusercontent.com/u/592928?v=4?s=100" width="100px;" alt=""/><br /><sub><b>Edward K. Ream</b></sub></a><br /><a href="#plugin-edreamleo" title="Plugin/utility libraries">🔌</a></td>\n    <td align="center"><a href="https://github.com/dastultz"><img src="https://avatars3.githubusercontent.com/u/4334042?v=4?s=100" width="100px;" alt=""/><br /><sub><b>Daryl Stultz</b></sub></a><br /><a href="#stubs-dastultz" title="MicroPython stubs">📚</a></td>\n    <td align="center"><a href="https://github.com/cabletie"><img src="https://avatars1.githubusercontent.com/u/2356734?v=4?s=100" width="100px;" alt=""/><br /><sub><b>Keeping things together</b></sub></a><br /><a href="https://github.com/Josverl/micropython-stubber/issues?q=author%3Acabletie" title="Bug reports">🐛</a></td>\n  </tr>\n  <tr>\n    <td align="center"><a href="https://github.com/vbolshakov"><img src="https://avatars2.githubusercontent.com/u/2453324?v=4?s=100" width="100px;" alt=""/><br /><sub><b>vbolshakov</b></sub></a><br /><a href="https://github.com/Josverl/micropython-stubber/issues?q=author%3Avbolshakov" title="Bug reports">🐛</a> <a href="#stubs-vbolshakov" title="MicroPython stubs">📚</a></td>\n    <td align="center"><a href="https://lemariva.com/"><img src="https://avatars2.githubusercontent.com/u/15173329?v=4?s=100" width="100px;" alt=""/><br /><sub><b>Mauro Riva</b></sub></a><br /><a href="#blog-lemariva" title="Blogposts">📝</a> <a href="https://github.com/Josverl/micropython-stubber/issues?q=author%3Alemariva" title="Bug reports">🐛</a></td>\n    <td align="center"><a href="https://github.com/MathijsNL"><img src="https://avatars0.githubusercontent.com/u/1612886?v=4?s=100" width="100px;" alt=""/><br /><sub><b>MathijsNL</b></sub></a><br /><a href="https://github.com/Josverl/micropython-stubber/issues?q=author%3AMathijsNL" title="Bug reports">🐛</a></td>\n    <td align="center"><a href="http://comingsoon.tm/"><img src="https://avatars0.githubusercontent.com/u/13251689?v=4?s=100" width="100px;" alt=""/><br /><sub><b>Callum Jacob Hays</b></sub></a><br /><a href="https://github.com/Josverl/micropython-stubber/issues?q=author%3ACallumJHays" title="Bug reports">🐛</a> <a href="#test-CallumJHays" title="Test">✔</a></td>\n    <td align="center"><a href="https://github.com/v923z"><img src="https://avatars0.githubusercontent.com/u/1310472?v=4?s=100" width="100px;" alt=""/><br /><sub><b>Zoltán Vörös</b></sub></a><br /><a href="#data-v923z" title="Data">🔣</a></td>\n    <td align="center"><a href="https://github.com/vincent-l-j"><img src="https://avatars.githubusercontent.com/u/20021376?v=4?s=100" width="100px;" alt=""/><br /><sub><b>vincent-l-j</b></sub></a><br /><a href="https://github.com/Josverl/micropython-stubber/commits?author=vincent-l-j" title="Documentation">📖</a></td>\n    <td align="center"><a href="https://github.com/yegorLitvinov"><img src="https://avatars.githubusercontent.com/u/20367310?v=4?s=100" width="100px;" alt=""/><br /><sub><b>Egor Litvinov</b></sub></a><br /><a href="https://github.com/Josverl/micropython-stubber/issues?q=author%3AyegorLitvinov" title="Bug reports">🐛</a></td>\n  </tr>\n</table>\n\n<!-- markdownlint-restore -->\n<!-- prettier-ignore-end -->\n\n<!-- ALL-CONTRIBUTORS-LIST:END -->\n\nThis project follows the [all-contributors](https://github.com/all-contributors/all-contributors) specification. Contributions of any kind welcome!\n\n----------------------------\n\n--------------------------------\n\n\n\n[stubs-repo]:   https://github.com/Josverl/micropython-stubs\n[stubs-repo2]:  https://github.com/BradenM/micropy-stubs\n[micropython-stubber]: https://github.com/Josverl/micropython-stubber\n[micropython-stubs]: https://github.com/Josverl/micropython-stubs#micropython-stubs\n[micropy-cli]: https://github.com/BradenM/micropy-cli\n[using-the-stubs]: https://github.com/Josverl/micropython-stubs#using-the-stubs\n[demo]:         https://github.com/Josverl/micropython-stubber/blob/main/docs/img/demo.gif?raw=true\t"demo of writing code using the stubs"\n[stub processing order]: https://github.com/Josverl/micropython-stubber/blob/main/docs/img/stuborder_pylance.png?raw=true\t"recommended stub processing order"\n[naming-convention]: #naming-convention-and-stub-folder-structure\n[all-stubs]: https://github.com/Josverl/micropython-stubs/blob/main/firmwares.md\n[micropython]: https://github.com/micropython/micropython\n[micropython-lib]:  https://github.com/micropython/micropython-lib\n[pycopy]: https://github.com/pfalcon/pycopy\n[pycopy-lib]: https://github.com/pfalcon/pycopy-lib\n\n',
    'author': 'Jos Verlinde',
    'author_email': 'jos_verlinde@hotmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Josverl/micropython-stubber#readme',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



"""
update-fallback folder with common set of stubs that cater for most of the devices
"""

import click
from pathlib import Path
from stubber.update_fallback import update_fallback, RELEASED
from stubber.utils.config import CONFIG

from .cli import stubber_cli

@stubber_cli.command(name="update-fallback")
@click.option("--version", default=RELEASED, type=str, help="Version number to use", show_default=True)
@click.option(
    "--stub-folder",
    default=CONFIG.stub_path.as_posix(),
    type=click.Path(exists=True, file_okay=False, dir_okay=True),
    help="Destination of the files to be generated.",
    show_default=True,
)
def cli_update_fallback(
    version: str,
    stub_folder: str = CONFIG.stub_path.as_posix(),
):
    """
    Update the fallback stubs.

    The fallback stubs are updated/collated from files of the firmware-stubs, doc-stubs and core-stubs.
    """
    stub_path = Path(stub_folder)
    update_fallback(
        stub_path,
        CONFIG.fallback_path,
        version=version,
    )



"""
ModuleSourceDict represents a source file with the following components 
    - docstr
    - version
    - comment
    - typing
    - Optional: list of constants
    - optional: ClassSourcedicts 
    - optional: FunctionSourcedicts 
    - optional: individual lines of code 

ClassSourceDict represents a source file with the following components 
    - comment
    - class
    - docstr
    - Optional: list of constants
    - __init__ : class signature 
    - optional: FunctionSourcedicts 
    - optional: individual lines of code 

FunctionSourceDict represents a source file with the following components 
    - # comments - todo
    - optional: decorator
    - def - function definition 
    - docstr
    - constants
    - body - ...
    - optional: individual lines of code 

SourceDict is the 'base class' 

"""
from __future__ import annotations

from typing import List, OrderedDict, Union

from .classsort import sort_classes

# These are shown to import
__all__ = [
    "SourceDict",
    "ModuleSourceDict",
    "ClassSourceDict",
    "FunctionSourceDict",
]


def spaces(n: int = 4) -> str:
    return " " * n


class SourceDict(OrderedDict):
    "(abstract) dict to store source components respecting parent child dependencies and proper definition order"

    def __init__(self, base: List, indent: int = 0, body: int = 0, lf: str = "\n"):
        super().__init__(base)
        self.lf = lf  #  add linefeed
        self.indent = indent  # current base indent
        self._body = body  # for source body is level
        self._nr = 0  # generate incrementing line numbers
        self.name = ""

    def __str__(self) -> str:
        "convert the OD into a string"
        out = ""
        for item in self.items():
            code = item[1]
            if isinstance(code, str):
                out += code + self.lf
            elif isinstance(code, List):
                for l in code:
                    if isinstance(l, str):
                        out += l + self.lf
                    else:
                        raise TypeError(f"Incorrect structure in Output dict: {l}") #noqa
            else:
                out += str(code)
        return out

    def __add__(self, dict: SourceDict):
        # sd = sd + function
        # sd += function
        self.update({dict.name: dict})
        return self

    def add_docstr(self, docstr: Union[str, List[str]], extra: int = 0):
        # indent +4,  add triple " to  docstring
        if isinstance(docstr, str):
            _docstr = [docstr]
        elif isinstance(docstr, List): #type: ignore
            _docstr = docstr.copy()
        else:
            raise TypeError
        if len(_docstr) > 0 and not _docstr[0].strip().startswith('"""'):
            # add triple quotes before & after
            quotes = '"""'
            _docstr.insert(0, quotes)
            _docstr.append(quotes)
        # add indent + extra
        _docstr = [spaces(self.indent + extra) + l for l in _docstr]
        self.update({"docstr": _docstr})

    def add_comment(self, line: Union[str, List[str]]):
        "Add a comment, or list of comments, to this block."
        _c = self["comment"] or []
        if isinstance(line, str):
            _c += [spaces(self.indent + self._body) + line]
        elif isinstance(line, list): #type: ignore
            for l in line:
                _c += [spaces(self.indent + self._body) + l]
        self.update({"comment": _c})

    def add_constant(self, line: str, autoindent: bool = True):
        "add constant to the constant scope of this block"
        if autoindent:
            line = spaces(self.indent + self._body) + line
        self.update({"constants": self["constants"] + [line]})

    def add_constant_smart(self, name: str, type: str = "Any", docstr: List[str] = [], autoindent: bool = True):
        """add constant to the constant scope of this block, or a class in this block"""
        if "." in name and isinstance(self, ModuleSourceDict):
            classname, const_name = name.split(".", 1)
            classfullname = self.find(classname.replace("# ", ""))  # Wildcards are
            if classfullname:
                cls_dict: ClassSourceDict = self[classfullname]
                cls_dict.add_constant_smart(const_name, type, docstr)
            else:
                raise KeyError(f"const {name} could not be added to Class {classname}")
        else:
            if "*" in name:
                # * wildcard used in documentation
                line = f"# {name}: {type}"
            else:
                line = f"{name}: {type}"
            # assign a value so constant can be used as default value
            if type == "Any":
                line += " = ..."
            elif type == "int":
                line += " = 1"
            elif type == "str":
                line += ' = ""'

            _docstr = docstr
            if autoindent:
                line = spaces(self.indent + self._body) + line
                if len(_docstr):
                    if len(_docstr) == 1:
                        #  - if len = 1 add triple quotes before & after, respecting indentaion
                        _docstr = [spaces(self.indent + self._body) + '"""' + _docstr[0].lstrip() + '"""']
                    else:
                        #  - if len > 1 add triple quotes on sep lines before & after,respecting indentaion
                        _docstr = (
                            [spaces(self.indent + self._body) + '"""\\']
                            + [spaces(self.indent + self._body) + l.lstrip() for l in _docstr]
                            + [spaces(self.indent + self._body) + '"""']
                        )
            if len(_docstr):
                #  - add docstring after defining constant
                self.update({"constants": self["constants"] + [line] + _docstr})
            else:
                self.update({"constants": self["constants"] + [line]})

    def find(self, name: str) -> Union[str, None]:
        raise NotImplementedError("Please Implement this method")

    def add_line(self, line: str, autoindent: bool = True):
        self._nr += 1
        if autoindent:
            line = spaces(self.indent + self._body) + line
        id = str(self._nr)
        self.update({id: line})
        return id

    def index(self, key: str):
        return list(self.keys()).index(key)


class ModuleSourceDict(SourceDict):
    def __init__(self, name: str, indent=0, lf: str = "\n"):
        "set correct order a module definition to allow adding class variables"
        super().__init__(
            [
                ("docstr", ['""" """']),
                ("version", ""),
                ("comment", []),
                ("imports", []),
                ("constants", []),
            ],
            indent,
            body=0,
            lf=lf,
        )
        self.name = name

    def sort(self):
        "make sure all classdefs are in order"
        # new empty one
        new = ModuleSourceDict(self.name, self.indent, self.lf)
        # add the standard stuff using a dict comprehension
        new.update(
            {
                k: v
                for (k, v) in self.items()
                if k
                in [
                    "docstr",
                    "version",
                    "comment",
                    "imports",
                    "constants",
                ]
            }
        )
        # then the classes, already sorted in parent-child order
        for classname in self.classes():
            new.update({classname: self[classname]})
        # then the functions and other
        new.update({k: v for (k, v) in self.items() if k.isdecimal() or k.startswith("def ")})
        # now clear and update with new order
        if len(self) != len(new):
            raise ValueError("Sort() changed the length of the dictionary")
        self.clear()
        self.update(new)

    def __str__(self):
        self.sort()
        return super().__str__()

    def find(self, name: str) -> Union[str, None]:
        "find a classnode based on the name with or without the superclass"
        keys = list(self.keys())
        if not name.startswith("class "):
            name = "class " + name
        # try full match first
        if name in keys:
            return name
        # is there a partial ? - only match before `(`
        name = name.split("(")[0]
        for k in keys:
            if name == k.split("(")[0]:
                return k
        return None

    def classes(self):
        "get a list of the class names in parent-child order"
        classes = [k for k in self.keys() if isinstance(self[k], ClassSourceDict)]
        # return sorted list
        return sort_classes(classes)

    def add_import(self, imports: Union[str, List[str]]):
        "add a [list of] imports this module"
        _imports = self["imports"] or []
        if isinstance(imports, str):
            imports = [imports]
        if "from __future__ import annotations" in imports:
            # from future ... must be the first import
            _imports = imports + _imports
        else:
            _imports += imports
        self.update({"imports": _imports})


class ClassSourceDict(SourceDict):
    def __init__(
        self,
        name: str,
        *,
        docstr: List[str] = ['""" """'],
        init: str = "",  # "def __init__(self)->None:",
        indent: int = 0,
        lf="\n",
    ):
        "set correct order for class and exception definitions to allow adding class variables"
        _init: List[str] = []
        if len(init) > 0:
            _init = [spaces(indent + 4) + init]
            # add ...
            _init.append(spaces(indent + 4 + 4) + "...")
        super().__init__(
            [
                ("comment", []),
                ("class", spaces(indent) + name),  # includes indentation
                ("docstr", ['""" """']),
                ("constants", []),
                ("__init__", _init),
            ],
            indent,
            body=4,  # class body  indent +4
            lf=lf,
        )
        self.add_docstr(docstr, extra=4)
        self.name = name
        self.lf = "\n"


class FunctionSourceDict(SourceDict):
    def __init__(
        self,
        name: str,
        *,
        definition: List[str] = [],
        docstr: List[str] = ['""" """'],
        indent: int = 0,
        decorators: List[str] = [],
        lf="\n",
    ):
        "set correct order for function and method definitions"
        # add indent
        _def = [spaces(indent) + l for l in definition]

        # add ...
        super().__init__(
            [
                ("decorator", [spaces(indent) + d for d in decorators]),
                ("def", _def),  # includes indentation
                ("docstr", '""'),  # just a placeholder
                #                ("comments", []),
                ("constants", []),
                ("body", spaces(indent + 4) + "..."),
            ],
            indent,
            body=4,  # function body indent +4
            lf=lf,
        )
        self.name = name
        self.add_docstr(docstr, extra=4)
        self.lf = "\n"



from .classsort import *
from .lookup import *
from .output_dict import *
from .rst_utils import *

# combine all _all__s
__all__ = classsort.__all__ + lookup.__all__ + output_dict.__all__ + rst_utils.__all__  # type: ignore
